#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ref = ot.SymbolicFunction("x", "sin(x)")
size = 12
locations = ot.Point(size)
values = ot.Point(size)
# Build locations/values with non-increasing locations
for i in range(size):
    locations[i] = 10.0 * i * i / (size - 1.0) / (size - 1.0)
    values[i] = ref([locations[i]])[0]

evaluation = ot.PiecewiseLinearEvaluation(locations, values)
print("evaluation=", evaluation)
# Check the values
for i in range(2 * size):
    x = [-1.0 + 12.0 * i / (2.0 * size - 1.0)]
    print("f( %.12g )=" % x[0], evaluation(x), ", ref=", ref(x))

# Test exception enableExtrapolation
locations = [1.0, 2.0, 3.0, 4.0, 5.0]
values = [-2.0, 2.0, 1.0, 3.0, 5.0]
evaluation = ot.PiecewiseLinearEvaluation(locations, values)
evaluation.setEnableExtrapolation(False)
f = ot.Function(evaluation)
with ott.assert_raises(TypeError):
    f([-12.5])



"""
Create a linear least squares model
===================================
"""

# %%
# In this example we are going to create a global approximation of a model
# response based on a linear function using the :class:`~openturns.LinearLeastSquares` class.
#
# We consider the function :math:`h : \Rset^2 \rightarrow \Rset^2` is defined by:
#
# .. math::
#    \vect{h}(x_1, x_2) = \Tr{\left(\cos(x_1 + x_2), (x_2 + 1) e^{x_1 - 2 x_2} \right)}
#
# for any :math:`\vect{x} \in \Rset^2`.
# Since the output is a dimension 2 vector, the model has vector
# coefficients.
# We use the linear model:
#
# .. math::
#     \vect{y} \, \approx \, \widehat{\vect{h}}(\vect{x}) \,
#     = \, \sum_{k=0}^2 \; \vect{a}_k \; \psi_k(\vect{x})
#
#
# for any :math:`\vect{x} \in \Rset^2` where :math:`\left\{\vect{a}_k \in \Rset^2\right\}_{k = 0,..., 2}`
# are the vector coefficients and :math:`\left\{\psi_k : \Rset^2 \rightarrow \Rset\right\}_{k = 0, ..., 2}`
# are the basis functions.
# This implies that each marginal output :math:`\widehat{h}_i(\vect{x})`
# is approximated by the linear model:
#
# .. math::
#     \widehat{h}_i(\vect{x}) \,
#     = \, \sum_{k=0}^2 \; a_{ki} \; \psi_k(\vect{x})
#
# for any :math:`\vect{x} \in \Rset^2` and any :math:`i = 1, 2`
# where :math:`a_{ki}` is the :math:`k`-th coefficient of the :math:`i`-th output
# marginal:
#
# .. math::
#      \vect{a}_k = \begin{pmatrix}a_{k1} \\ a_{k2} \end{pmatrix}
#
# for :math:`k = 0, 1, 2`.
# We consider the basis functions:
#
# .. math::
#
#      \psi_0(\vect{x}) & = 1 \\
#      \psi_1(\vect{x}) & = x_1 \\
#      \psi_2(\vect{x}) & = x_2 \\
#
# for any :math:`\vect{x} \in \Rset^2`.


# %%
import openturns as ot
import openturns.viewer as otv


# %%
# Define the model
# ~~~~~~~~~~~~~~~~

# %%
# Prepare an input sample.
# Each point is a pair of coordinates :math:`(x_1, x_2)`.

# %%
inputTrain = [[0.5, 0.5], [-0.5, -0.5], [-0.5, 0.5], [0.5, -0.5]]
inputTrain += [[0.25, 0.25], [-0.25, -0.25], [-0.25, 0.25], [0.25, -0.25]]
inputTrain = ot.Sample(inputTrain)
inputTrain.setDescription(["x1", "x2"])
inputTrain

# %%
# Compute the output sample from the input sample and a function.

# %%
formulas = ["cos(x1 + x2)", "(x2 + 1) * exp(x1 - 2 * x2)"]
model = ot.SymbolicFunction(["x1", "x2"], formulas)
model.setOutputDescription(["y1", "y2"])
outputTrain = model(inputTrain)
outputTrain

# %%
# Linear least squares
# ~~~~~~~~~~~~~~~~~~~~

# %%
# Create a linear least squares model.

# %%
algo = ot.LinearLeastSquares(inputTrain, outputTrain)
algo.run()

# %%
# Get the linear term.

# %%
algo.getLinear()

# %%
# Get the constant term.

# %%
algo.getConstant()

# %%
# Get the metamodel.
responseSurface = algo.getResult().getMetaModel()

# %%
# Plot the second output of our model with :math:`x_1=0.5`.

# %%
graph = ot.ParametricFunction(model, [0], [0.5]).getMarginal(1).draw(-0.5, 0.5)
graph.setLegends(["Model"])
curve = (
    ot.ParametricFunction(responseSurface, [0], [0.5])
    .getMarginal(1)
    .draw(-0.5, 0.5)
    .getDrawable(0)
)
curve.setLineStyle("dashed")
curve.setLegend("Linear L.S.")
graph.add(curve)
graph.setLegendPosition("upper right")
view = otv.View(graph)

# %%
view.ShowAll()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


center = ot.Point(4, 0.0)
center.setName("center")
center[0] = 0.5
center[1] = 1.5
center[2] = 2.5
center[3] = 3.5
levels = ot.Point(3, 0.0)
levels.setName("levels")
levels[0] = 4
levels[1] = 8
levels[2] = 16
experiment = ot.Factorial(center, levels)
print("experiment = ", experiment)
sample = experiment.generate()
print("sample = ", repr(sample))



"""
Plot the log-likelihood contours of a distribution
==================================================
"""

# %%
# In this example, we show how to plot the bidimensionnal log-likelihood contours of function given a sample.

# %%
import openturns.viewer as otv
import openturns as ot

# %%
# Generate a sample
# -----------------

# %%
# We create a :class:`~openturns.TruncatedNormal` and generate a small sample.

# %%
a = -1
b = 2.5
mu = 2.0
sigma = 3.0
distribution = ot.TruncatedNormal(mu, sigma, a, b)
sample = distribution.getSample(11)

# %%
# In order to see the distribution and the sample, we draw the PDF of the distribution and generate a cloud which `X` coordinates are the sample values.

# %%
graph = distribution.drawPDF()
graph.setLegends(["TruncatedNormal"])
zeros = ot.Sample(sample.getSize(), 1)
cloud = ot.Cloud(sample, zeros)
cloud.setLegend("Sample")
graph.add(cloud)
graph.setLegendPosition("upper left")
view = otv.View(graph)


# %%
# The following function computes the log-likelihood of a :class:`~openturns.TruncatedNormal`
# which mean and standard deviations are given as input arguments.
# The lower and upper bounds of the distribution are computed as minimum and maximum of the sample.

# %%
# Define the log-likelihood function
# ----------------------------------

# %%
# The following function evaluates the log-likelihood function given a point :math:`X=(\mu,\sigma`).
# In order to evaluate the likelihood on the sample, we use a trick: we evaluate
# the `computeMean` method on the log-PDF sample, then multiply by the sample size.
# This is much faster than using a `for` loop.


# %%
def logLikelihood(X):
    """
    Evaluate the log-likelihood of a TruncatedNormal on a sample.
    """
    samplesize = sample.getSize()
    mu = X[0]
    sigma = X[1]
    a = sample.getMin()[0]
    b = sample.getMax()[0]
    delta = (b - a) / samplesize
    a -= delta
    b += delta
    distribution = ot.TruncatedNormal(mu, sigma, a, b)
    samplelogpdf = distribution.computeLogPDF(sample)
    loglikelihood = samplelogpdf.computeMean() * samplesize
    return loglikelihood


# %%
# With the draw method
# --------------------

# %%
# In this section, we use the `draw` method which is available for any `Function` which has 1 or 2 input arguments.
# In our case, the log-likelihood function has two inputs: :math:`x_0=\mu` and :math:`x_1=\sigma`.

# %%
# Draw the log-likelihood function with the `draw` method: this is much faster than using a `for` loop.
# In order to print LaTeX `X` and `Y` labels, we use the `"r"` character in front of the string containing the LaTeX command.

# %%
logLikelihoodFunction = ot.PythonFunction(2, 1, logLikelihood)
graphBasic = logLikelihoodFunction.draw([-3.0, 0.1], [5.0, 7.0], [50] * 2)
graphBasic.setXTitle(r"$\mu$")
graphBasic.setYTitle(r"$\sigma$")
view = otv.View(graphBasic)

# %%
# Customizing the number of levels
# --------------------------------

# %%
# The level values are computed from the quantiles of the data, so that the contours are equally spaced.
# We can configure the number of levels by setting the `Contour-DefaultLevelsNumber` key in the :class:`~openturns.ResourceMap`.

# %%
ot.ResourceMap.SetAsUnsignedInteger("Contour-DefaultLevelsNumber", 5)
logLikelihoodFunction = ot.PythonFunction(2, 1, logLikelihood)
graphBasic = logLikelihoodFunction.draw([-3.0, 0.1], [5.0, 7.0], [50] * 2)
graphBasic.setXTitle(r"$\mu$")
graphBasic.setYTitle(r"$\sigma$")
view = otv.View(graphBasic)

# %%
# We get the underlying `Contour` object as a `Drawable`.

contour = graphBasic.getDrawable(0)

# %%
# To be able to use specific `Contour` methods like `buildDefaultLevels`, we need to use the method named `getImplementation`.

contour = contour.getImplementation()
contour.buildDefaultLevels(50)

manyLevelGraph = ot.Graph()
manyLevelGraph.add(contour)
view = otv.View(manyLevelGraph)

# %%
# Using a rank-based normalization of the colors
# ----------------------------------------------

# %%
# In the previous plots, there was little color variation for isolines corresponding to large log-likelihood values.
# This is due to a steep cliff visible for low values of :math:`\sigma`.
# To make the color variation clearer around -13, we use a normalization based on the rank of the level curve and not on its value.
contour.setColorMapNorm("rank")
rankGraph = ot.Graph()
rankGraph.add(contour)
view = otv.View(rankGraph)

# %%
# Fill the contour graph
# ----------------------

# %%
# Areas between contour lines can be colored by requesting a filled outline.

# sphinx_gallery_thumbnail_number = 6
contour.setIsFilled(True)
filledGraph = ot.Graph()
filledGraph.add(contour)
view = otv.View(filledGraph)

# %%
# Getting the level values
# ------------------------

# %%
# The level values can be retrieved with the `getLevels` method.

# %%
drawables = graphBasic.getDrawables()
levels = drawables[0].getLevels()
levels

# %%
# Monochrome contour plot
# -----------------------

# %%
# We first configure the contour plot.
# We use the `getDrawable` method to take the first contour as the only one with multiple levels.
# Then we use the `setLevels` method: we could have changed the levels.
# We use the `setColor` method to get a monochrome contour.
# In order to inline the level values labels, we use the `setDrawLabels` method.

# %%
contour = graphBasic.getDrawable(0)
contour.setLevels(levels)
contour.setDrawLabels(True)
contour.setColor("red")

# %%
# Hide the color bar.
# The method to do this is not available to generic Drawables,
# so we need to get the actual `Contour` object.

contour = contour.getImplementation()
contour.setColorBarPosition("")

# %%
# Then we create a new graph. Finally, we use `setDrawables` to define this `Contour` object as the single Drawable.

graphFineTune = ot.Graph("Log-Likelihood", r"$\mu$", r"$\sigma$")
graphFineTune.setDrawables([contour])
view = otv.View(graphFineTune)

# %%
# Set multiple colors manually
# ----------------------------

# %%
# The :class:`~openturns.Contour` class does not allow us to manually set multiple colors.
# Here we show how to assign explicit colors to the different contour lines by passing keyword
# arguments to the class:`~openturns.viewer.View` class.

# Build a range of colors corresponding to the Tableau palette
palette = ot.Drawable.BuildTableauPalette(len(levels))
view = otv.View(graphFineTune, contour_kw={"colors": palette, "cmap": None})

# %%
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot

# Tests
ot.ResourceMap.SetAsUnsignedInteger("SimulationAlgorithm-DefaultBlockSize", 100)
ot.ResourceMap.SetAsUnsignedInteger(
    "SimulationAlgorithm-DefaultMaximumOuterSampling", 100
)
ot.ResourceMap.SetAsScalar(
    "SimulationAlgorithm-DefaultMaximumCoefficientOfVariation", 0.0
)
ot.ResourceMap.SetAsScalar("SimulationAlgorithm-DefaultMaximumStandardDeviation", 0.0)
ot.ResourceMap.SetAsScalar("RootStrategy-DefaultStepSize", 0.1)

algorithms = ["MonteCarlo", "LHS", "QuasiMonteCarlo", "DirectionalSampling"]

inDim = 4
X = ot.RandomVector(ot.Normal(inDim))
inVars = ot.Description.BuildDefault(inDim, "x")

low = 1.0
up = 2.0
intervals = [
    ot.Interval([low], [up], [True], [False]),
    ot.Interval([low], [up], [False], [True]),
    ot.Interval([low], [up], [True], [True]),
    ot.Interval([low], [up], [False], [False]),
    ot.Interval([low] * 2, [up] * 2, [True, True], [True, True]),
    ot.Interval([low] * 2, [up] * 2, [True, True], [True, False]),
    ot.Interval([low] * 2, [up] * 2, [True, True], [False, True]),
    ot.Interval([low] * 2, [up] * 2, [True, True], [False, False]),
    ot.Interval([low] * 2, [up] * 2, [True, False], [True, True]),
    ot.Interval([low] * 2, [up] * 2, [True, False], [True, False]),
    ot.Interval([low] * 2, [up] * 2, [True, False], [False, True]),
    ot.Interval([low] * 2, [up] * 2, [True, False], [False, False]),
    ot.Interval([low] * 2, [up] * 2, [False, True], [True, True]),
    ot.Interval([low] * 2, [up] * 2, [False, True], [True, False]),
    ot.Interval([low] * 2, [up] * 2, [False, True], [False, True]),
    ot.Interval([low] * 2, [up] * 2, [False, True], [False, False]),
    ot.Interval([low] * 2, [up] * 2, [False, False], [True, True]),
    ot.Interval([low] * 2, [up] * 2, [False, False], [True, False]),
    ot.Interval([low] * 2, [up] * 2, [False, False], [False, True]),
    ot.Interval([low] * 2, [up] * 2, [False, False], [False, False]),
]

for domain in intervals:
    print("#" * 50)
    print("domain=\n", domain)
    outDim = domain.getDimension()
    f = ot.SymbolicFunction(inVars, inVars[0:outDim])
    Y = ot.CompositeRandomVector(f, X)
    event = ot.ThresholdEvent(Y, domain)

    ot.RandomGenerator.SetSeed(0)
    algo = ot.ProbabilitySimulationAlgorithm(event, ot.MonteCarloExperiment())
    algo.run()
    res = algo.getResult().getProbabilityEstimate()
    print("MC p=%.6g" % res)

    ot.RandomGenerator.SetSeed(0)
    algoOptim = ot.Cobyla()
    algoOptim.setStartingPoint(X.getMean())
    algoOptim.setMaximumConstraintError(1)
    algo = ot.FORM(algoOptim, event)

    algo.run()
    res = algo.getResult().getEventProbability()
    print("FORM p=%.2f" % res)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


# Default dimension parameter to evaluate the model
defaultDimension = 1

# Amplitude values
amplitude = ot.Point(defaultDimension, 1.0)

# Scale values
scale = ot.Point(defaultDimension, 1.0)

# Default constructor
myDefaultModel = ot.CauchyModel()
print("myDefaultModel = ", myDefaultModel)

# Second order model with parameters
myModel = ot.CauchyModel(scale, amplitude)
print("myModel = ", myModel)

frequencyValue = 1.0
print("spectral density matrix at f = ", frequencyValue, " : ", myModel(frequencyValue))

# Evaluation at time higher to check the decrease of the cauchy values
frequencyValueHigh = 10.0
print(
    "spectral density matrix at f = ",
    frequencyValueHigh,
    " : ",
    myModel(frequencyValueHigh),
)

# Default dimension parameter to evaluate the model
highDimension = 3

# Reallocation of adequate sizes
amplitude.resize(highDimension)
scale.resize(highDimension)
spatialCorrelation = ot.CorrelationMatrix(highDimension)
for index in range(highDimension):
    amplitude[index] = 1.0
    scale[index] = (index + 1.0) / (defaultDimension * defaultDimension)
    if index > 0:
        spatialCorrelation[index, index - 1] = 1.0 / (index * index)

# check the cast
mySecondOrderModel = ot.SpectralModel(
    ot.CauchyModel(scale, amplitude, spatialCorrelation)
)
print("mySecondOrderModel = ", mySecondOrderModel)
# checking the cast

# Second order model  - dimension 10
myHighModel = ot.CauchyModel(scale, amplitude, spatialCorrelation)
print("myHighModel = ", myHighModel)
print("spectral density matrix at f = ", frequencyValue, " : ", myModel(frequencyValue))
print(
    "spectral density matrix at f = ",
    frequencyValueHigh,
    " : ",
    myModel(frequencyValueHigh),
)



#! /usr/bin/env python

import openturns as ot
import os

ot.TESTPREAMBLE()

nrVertices = 100
vertices = ot.Normal().getSample(nrVertices).sort()
simplices = [[i, i + 1] for i in range(nrVertices - 1)]
mesh1 = ot.Mesh(vertices, simplices)
vertices *= -1.0
mesh2 = ot.Mesh(vertices, simplices)

for mesh in [mesh1, mesh2]:
    lowerBound = mesh.getLowerBound()[0]
    upperBound = mesh.getUpperBound()[0]
    n = mesh.getSimplicesNumber()
    print(
        "mesh=",
        mesh,
        "lowerBound=",
        lowerBound,
        "upperBound=",
        upperBound,
        n,
        "simplices",
    )
    algo = ot.EnclosingSimplexMonotonic1D(mesh.getVertices())

    ot.RandomGenerator.SetSeed(0)
    test = ot.Sample(ot.Uniform(-3.0, 3.0).getSample(1000))

    vertices = mesh.getVertices()
    for vertex in test:
        index = algo.query(vertex)
        x = vertex[0]
        if x < lowerBound or x > upperBound:
            if index < n:
                print("Point", x, "should be outside but query returned index", index)
                os.exit(1)
        else:
            if index >= n:
                print("Point", x, "should be inside but query returned index", index)
                os.exit(1)

            found, coordinates = mesh.checkPointInSimplexWithCoordinates(vertex, index)
            if not found:
                print(
                    "Wrong simplex found for",
                    x,
                    "(index=",
                    index,
                    ") barycentric coordinates=",
                    coordinates,
                )
                os.exit(1)
            if (
                coordinates[0] < 0.0
                or coordinates[0] > 1.0
                or coordinates[1] < 0.0
                or coordinates[1] > 1.0
            ):
                print(
                    "Wrong barycentric coordinates found found for",
                    x,
                    "(index=",
                    index,
                    ") barycentric coordinates=",
                    coordinates,
                )
                os.exit(1)
            if (
                abs(
                    x
                    - coordinates[0] * vertices[index, 0]
                    - coordinates[1] * vertices[index + 1, 0]
                )
                > 1.0e-10
            ):
                print(
                    "Wrong barycentric coordinates found found for",
                    x,
                    "(index=",
                    index,
                    ") barycentric coordinates=",
                    coordinates,
                )
                os.exit(1)



#! /usr/bin/env python

import openturns as ot
import multiprocessing


ot.TESTPREAMBLE()


def myfunPython(x):
    n = x[0]
    y = [1.0 / (2.0**n)]
    return y


if __name__ == "__main__":
    n_cpus = multiprocessing.cpu_count()
    nsim = 1000
    myfun = ot.PythonFunction(1, 1, myfunPython, n_cpus=n_cpus)

    X = [[i] for i in range(nsim)]
    Y = myfun(X)
    sum_y = nsim * Y.computeMean()[0]
    assert sum_y == 2.0, "sum 2^-n = 2"



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


histogramPolynomial = ot.HistogramPolynomialFactory(
    -2.0, [1.0, 2.0, 1.5], [2.0, 5.0, 3.5]
)
print("histogramPolynomial=", histogramPolynomial)
for i in range(10):
    print("histogramPolynomial(", i, ")=", histogramPolynomial.build(i))
roots = histogramPolynomial.getRoots(10)
print("histogramPolynomial(10) roots=", repr(roots))
nodes, weights = histogramPolynomial.getNodesAndWeights(10)
print("histogramPolynomial(10) nodes=", nodes, "and weights=", weights)



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

levelFunction = ot.SymbolicFunction(["x1", "x2", "x3", "x4"], ["x1+2*x2-3*x3+4*x4"])
# Add a finite difference gradient to the function, as Abdo Rackwitz algorithm
# needs it
myGradient = ot.NonCenteredFiniteDifferenceGradient(1e-7, levelFunction.getEvaluation())
print("myGradient = ", repr(myGradient))
# Substitute the gradient
levelFunction.setGradient(ot.NonCenteredFiniteDifferenceGradient(myGradient))
startingPoint = [0.0] * 4
algo = ot.AbdoRackwitz(ot.NearestPointProblem(levelFunction, 3.0))
algo.setStartingPoint(startingPoint)
algo.run()
print("result = ", algo.getResult())
ott.assert_almost_equal(algo.getResult().getOptimalValue(), [3.0])
levelFunction = ot.SymbolicFunction(
    ["x1", "x2", "x3", "x4"], ["x1*cos(x1)+2*x2*x3-3*x3+4*x3*x4"]
)
# Add a finite difference gradient to the function, as Abdo Rackwitz algorithm
# needs it
myGradient = ot.NonCenteredFiniteDifferenceGradient(1e-7, levelFunction.getEvaluation())
print("myGradient = ", repr(myGradient))
# Substitute the gradient
levelFunction.setGradient(ot.NonCenteredFiniteDifferenceGradient(myGradient))
startingPoint = [0.0] * 4
algo = ot.AbdoRackwitz(ot.NearestPointProblem(levelFunction, -0.5))
algo.setStartingPoint(startingPoint)
print("myalgorithm=", repr(algo))
algo.run()
print("result = ", algo.getResult())
ott.assert_almost_equal(algo.getResult().getOptimalValue(), [-0.5])
print("evaluation calls number=", levelFunction.getEvaluationCallsNumber())
print("gradient   calls number=", levelFunction.getGradientCallsNumber())
print("hessian    calls number=", levelFunction.getHessianCallsNumber())



"""
Estimate a GEV on the Venice sea-levels data
============================================
"""

# %%
# In this example, we illustrate various techniques of extreme value modeling applied
# to the annual maximum sea-levels recorded in Venice over the period 1931-1981.
# Readers should refer to [coles2001]_ to get more details.
#
# We illustrate techniques to:
#
# - estimate a stationary GEV, using the :math:`r`-largest annual sea-levels for :math:`r\geq 1`,
#
# using:
#
# - the log-likelihood function,
# - the profile log-likelihood function.
#
# First, we load the 10 largest annual sea-levels in Venice. We start by looking at them
# through time. Note that for the year 1935, only the largest 6 observations are available.
# For simplicity of the example, we removed that year from the data but it could be
# used for all the analyses beased on the largest :math:`r` annual sea-levels for :math:`r \leq 6`.
import openturns as ot
import openturns.viewer as otv
from openturns.usecases import coles

data = coles.Coles().venice

# %%
# The column :math:`i` contains the largest :math:`i` annual sea-levels.
print(data[:5])
graph = ot.Graph("Annual maximum sea-levels in Venice", "year", "level (cm)")
for r in range(10):
    cloud = ot.Cloud(data[:, [0, 1 + r]])
    graph.add(cloud)
graph.setIntegerXTick(True)
view = otv.View(graph)

# %%
# **Stationary GEV modeling from the annual maximum sea-levels**
#
# We first assume that the dependence through time is negligible, so we first model the data as
# independent observations over the observation period. We estimate the parameters of the
# GEV distribution by maximizing the log-likelihood of the data.
# We select the first column of the data.
sample = data[:, 1]

# %%
# Estimate the parameters of the GEV by maximizing the log-likehood
# and compute the parameter distribution
factory = ot.GeneralizedExtremeValueFactory()
result_LL_max = factory.buildMethodOfLikelihoodMaximizationEstimator(sample)

# %%
# We get the fitted GEV and its parameters of :math:`(\hat{\mu}, \hat{\sigma}, \hat{\xi})`.
fitted_GEV = result_LL_max.getDistribution()
desc = fitted_GEV.getParameterDescription()
param = fitted_GEV.getParameter()
print(", ".join([f"{p}: {value:.3f}" for p, value in zip(desc, param)]))
print("Max log-likelihood (one max): ", result_LL_max.getLogLikelihood())

# %%
# We get the asymptotic distribution of the estimator :math:`(\hat{\mu}, \hat{\sigma}, \hat{\xi})`.
# In that case, the asymptotic distribution is normal.
parameterEstimate = result_LL_max.getParameterDistribution()
print("Asymptotic distribution of the estimator : ")
print(parameterEstimate)

# %%
# We get the covariance matrix  and the standard deviation of :math:`(\hat{\mu}, \hat{\sigma}, \hat{\xi})`.
print("Cov matrix = ", parameterEstimate.getCovariance())
print("Standard dev = ", parameterEstimate.getStandardDeviation())

# %%
# We get the marginal confidence intervals of order 0.95.
order = 0.95
for i in range(3):
    ci = parameterEstimate.getMarginal(i).computeBilateralConfidenceInterval(order)
    print(desc[i] + ":", ci)

# %%
# At last, we can validate the inference result thanks the 4 usual diagnostic plots:
#
# - the probability-probability pot,
# - the quantile-quantile pot,
# - the return level plot,
# - the empirical distribution function.
validation = ot.GeneralizedExtremeValueValidation(result_LL_max, sample)
graph = validation.drawDiagnosticPlot()
view = otv.View(graph)

# %%
# We can also use the profile log-likehood function rather than log-likehood function  to estimate the parameters of the GEV.
result_PLL_max = factory.buildMethodOfXiProfileLikelihoodEstimator(sample)

# %%
# The following graph allows one to get the profile log-likelihood plot.
# It also indicates the optimal value of :math:`\xi`, the maximum profile log-likelihood and
# the confidence interval for :math:`\xi` of order 0.95 (which is the default value).
order = 0.95
result_PLL_max.setConfidenceLevel(order)
print(result_PLL_max.getParameterConfidenceInterval())

# %%
# We can get the numerical values of the confidence interval: it appears to be a bit smaller
# with the interval obtained from the profile log-likelihood function than with the log-likelihood function.
# Note that if the order requested is too high, the confidence interval might not be calculated because
# one of its bound is out of the definition domain of the log-likelihood function.
try:
    print(
        "Confidence interval for xi = ", result_PLL_max.getParameterConfidenceInterval()
    )
except Exception as ex:
    print(type(ex))
    pass

# %%
# **Stationary GEV modeling from the largest** :math:`r` **annual sea-levels**
#
# We still assume that the dependence through time is negligible.  We estimate the parameters of the
# GEV distribution by maximizing the log-likelihood of the data. Now, we want to model more of the
# observed extremes than the annual maxima: the additional information contained in the largest
# :math:`10` observations can be used to improve the estimation of the GEV model.
#
# Now, we drop the year column to keep only the maxima values.
sample_rmax = data[:, 1:]
print(sample_rmax[:5])

# %%
# We estimate the parameters from the largest :math:`r` annual sea-levels for :math:`r=1,5,10`.
# For each :math:`r` value, we get the estimated parameters.
factory = ot.GeneralizedExtremeValueFactory()
r_candidate = [1, 5, 10]
for r in r_candidate:
    estimate = factory.buildMethodOfLikelihoodMaximization(sample_rmax, r)
    desc = estimate.getParameterDescription()
    p = estimate.getParameter()
    pretty_p = ", ".join([f"{param}: {value:.3f}" for param, value in zip(desc, p)])
    print(f"r={r:2} {pretty_p}")

# %%
otv.View.ShowAll()



"""
Create a Gaussian process with or without nugget effect
=======================================================
"""

# %%
# This example shows how to create a Gaussian process:
#
# - from its covariance function,
# - from its covariance function including a nugget effect,
# - from its spectral density in the stationary case.

# %%
import openturns as ot
import openturns.viewer as otv

# sphinx_gallery_thumbnail_number = 2

# %%
# Create a Gaussian process from its covariance model
# ---------------------------------------------------
#
# Here, we create a Gaussian process :math:`\vect{X}: \Omega \times \cD \rightarrow \Rset` on a
# domain :math:`\cD \in \Rset` from its covariance model.
#
# See :ref:`Covariance models <covariance_model>` to get more details on covariance models.
#
# We first define a covariance model. We use a :class:`~openturns.MaternModel` model. By default, the nugget factor
# is equal to :math:`10^{-12}` and is defined through the `CovarianceModel-DefaultNuggetFactor` key in the class
# :class:`~openturns.ResourceMap`. That is why we set it to zero.
amplitude = [1.0]
scale = [1.0]
myModel = ot.MaternModel(scale, amplitude, 2.5)
myModel.setNuggetFactor(0.0)

# %%
# We define a mesh that discretizes the domain :math:`\cD`. We use a  :class:`~openturns.RegularGrid`.
tmin = 0.0
N = 200
tmax = 5
step = (tmin - tmax) / (N - 1)
myTimeGrid = ot.RegularGrid(tmin, step, N)

# %%
# We create the Gaussian process from the covariance model and the mesh.
process = ot.GaussianProcess(myModel, myTimeGrid)
print(process)

# %%
# We draw some realizations of the Gaussian process.
n_real = 3
sample = process.getSample(n_real)
graph = sample.drawMarginal(0)
graph.setTitle("Processs realizations")
view = otv.View(graph)


# %%
# Add a nugget effect to the Gaussian process
# -------------------------------------------
#
# Here, we add a nugget effect to the Gaussian process. We use the previous Gaussian process. The nugget
# factor is
# :math:`\varepsilon_{nugget}`.
#
# Refer to :ref:`Covariance models <covariance_model>` to get more details on covariance models and the
# introduction of a
# nugget factor, and in particular see equation :eq:`Css`.
#
# Refer also to the example :doc:`/auto_stochastic_processes/plot_covariance_model_nugget` which illustrates
# the impact of the nugget effect on a covariance model.
#
# Adding a nugget factor modifies the covariance model. It transforms the process by adding a white noise
# of dimension :math:`d` with zero mean and a covariance matrix equal to
# :math:`\varepsilon_{nugget} \mat{C}^{spatial}`:
#
# .. math::
#
#   \vect{X}_{nugget}(\omega, \vect{t}) = \vect{X}(\omega, \vect{t}) +
#   \vect{\varepsilon}(\omega), \quad \vect{\varepsilon} \sim \cN(\vect{0},
#   \varepsilon_{nugget} \mat{C}^{spatial})
#
# We fix :math:`\varepsilon_{nugget} = 0.05`.
epsilon_nugget = 0.05
myModel.setNuggetFactor(epsilon_nugget)
process_nugget = ot.GaussianProcess(myModel, myTimeGrid)
print(process_nugget)

# %%
# We draw some realizations of the Gaussian process. We notice that the
# realizations of the process with the nugget factor are more chaotic than the other ones.
sample_nugget = process_nugget.getSample(n_real)
graph_nugget = sample_nugget.drawMarginal(0)
graph_nugget.setTitle("Process realizations with nugget effect")
view = otv.View(graph_nugget)

# %%
# Create a Gaussian process from its spectral density
# ---------------------------------------------------
#
# Here, we create a Gaussian process :math:`\vect{X}: \Omega \times \cD \rightarrow \Rset` on a
# domain :math:`\cD \in \Rset` from its spectral density.
#
# We first define a spectral model. We use a :class:`~openturns.CauchyModel`.
amplitude = [1.0]
scale = [4.0]
mySpectralModel = ot.CauchyModel(scale, amplitude)

# %%
# We define the associated mesh which is a :class:`~openturns.RegularGrid`.
myTimeGrid = ot.RegularGrid(0.0, 0.1, 20)

# %%
# We create the Gaussian process using a :class:`~openturns.SpectralGaussianProcess`.
process = ot.SpectralGaussianProcess(mySpectralModel, myTimeGrid)
print(process)

# %%
# Eventually we draw some realizations of the Gaussian process.
sample = process.getSample(n_real)
graph = sample.drawMarginal(0)
graph.setTitle("Processs realizations")
view = otv.View(graph)

# %%
# Display all figures.
otv.View.ShowAll()



"""
Gaussian Process Regression: multiple input dimensions
======================================================
"""

# %%
# In this example we are going to create an approximation of a model response using a Gaussian Process Regression
# model (refer to :ref:`gaussian_process_regression`).
# We consider a bidimensional function with Gaussian inputs.
# Then we create a GP metamodel with a constant basis and a :class:`~openturns.SquaredExponential` covariance.
#
# We consider the function
#
# .. math::
#    g(\vect{x}) = \cos(x_1 + x_2)
#
#
# for any :math:`\vect{x} \in \Rset^2`.
# We assume that :math:`X_1` and :math:`X_2` have a Gaussian distribution:
#
# .. math::
#    X_1 \sim \mathcal{N}(0,1) \textrm{ and } X_2 \sim \mathcal{N}(0,1).
#

# %%
import openturns as ot
import openturns.viewer as otv


# %%
# We define the model.

# %%
dimension = 2
input_names = ["x1", "x2"]
formulas = ["cos(x1 + x2)"]
model = ot.SymbolicFunction(input_names, formulas)

# %%
# We generate a simple Monte-Carlo input sample and evaluate the corresponding output sample.

# %%
distribution = ot.Normal(dimension)
samplesize = 15
x = distribution.getSample(samplesize)
y = model(x)

# %%
# Then we create a GP metamodel, using a constant trend and a squared exponential covariance model.
basis = ot.ConstantBasisFactory(dimension).build()
covarianceModel = ot.SquaredExponential([0.1] * dimension, [1.0])
fitter = ot.GaussianProcessFitter(x, y, covarianceModel, basis)
fitter.run()
fitter_result = fitter.getResult()
algo = ot.GaussianProcessRegression(fitter_result)
algo.run()
result = algo.getResult()
metamodel = result.getMetaModel()

# %%
# It is not so easy to visualize a bidimensional function. In order to simplify the graphics, we consider the value of the function at the input :math:`x_{1,ref}=0.5`.
# This amounts to create a :class:`~openturns.ParametricFunction` where the first variable :math:`x_1` (at input index 0) is set to :math:`0.5`.

# %%
x1ref = 0.5
metamodelAtXref = ot.ParametricFunction(metamodel, [0], [x1ref])
modelAtXref = ot.ParametricFunction(model, [0], [x1ref])

# %%
# For this given value of :math:`x_1`, we plot the model and the metamodel with :math:`x_2` from its 1% up to its 99% quantile.
# We configure the X title to "X2" because the default setting would state that this axis is the first value of the parametric function, which default name is "X0".

# %%
x2min = ot.Normal().computeQuantile(0.01)[0]
x2max = ot.Normal().computeQuantile(0.99)[0]
graph = metamodelAtXref.draw(x2min, x2max)
graph.setLegends(["Kriging"])
curve = modelAtXref.draw(x2min, x2max)
curve.setLegends(["Model"])
curve.setColors(["red"])
graph.add(curve)
graph.setLegendPosition("upper right")
graph.setTitle("Sample size = %d" % (samplesize))
graph.setXTitle("X2")
view = otv.View(graph)

# %%
view.ShowAll()

# %%
# As we can see, the metamodel is quite accurate in this case.
# The metamodel is very close to the model in the center of the domain, where the density of the input distribution is highest.



"""
Multi-output Gaussian Process Regression on the fire satellite model
====================================================================
"""

# %%
# This example aims to illustrate Gaussian Process Fitter (Kriging) metamodel with several outputs on the fire satellite model.


# %%
# Loading of the model
# --------------------
# This model involves 9 input variables and 3 output variables.
# We load the :ref:`Fire satellite use case<use-case-fire-satellite>`.

# %%
import openturns as ot
from openturns.usecases.fire_satellite import FireSatelliteModel
import openturns.viewer as otv

# %%
# We define the function that evaluates the outputs depending on the inputs.
m = FireSatelliteModel()
model = m.model

# %%
# We also define the distribution of input variables to build the training and test sets.
inputDistribution = m.distribution


# %%
# Generation of data
# ------------------
# We now generate the input and output training sets as 10 times the dimension of the input vector.

# %%
experiment = ot.LHSExperiment(inputDistribution, 10 * m.dim)
inputTrainingSet = experiment.generate()
outputTrainingSet = model(inputTrainingSet)

print("Lower and upper bounds of inputTrainingSet:")
print(inputTrainingSet.getMin(), inputTrainingSet.getMax())

# %%
# Creation of metamodel
# ---------------------
# We choose to use a constant trend.

# %%
linear_basis = ot.LinearBasisFactory(m.dim).build()
basis = ot.Basis(
    [
        ot.AggregatedFunction([linear_basis.build(k)] * 3)
        for k in range(linear_basis.getSize())
    ]
)

# %%
# We would like to have separate covariance models for the three outputs.
# To do so, we use the :class:`~openturns.TensorizedCovarianceModel`.
# For the purpose of illustration, we consider :class:`~openturns.MaternModel` for the first and third outputs, and :class:`~openturns.SquaredExponential` for the second output.

# %%
myCov1 = ot.MaternModel([1.0] * m.dim, 2.5)
myCov2 = ot.SquaredExponential([1.0] * m.dim)
myCov3 = ot.MaternModel([1.0] * m.dim, 2.5)
covarianceModel = ot.TensorizedCovarianceModel([myCov1, myCov2, myCov3])

# %%
# We can now define the GP fitter model.
fitter_algo = ot.GaussianProcessFitter(
    inputTrainingSet, outputTrainingSet, covarianceModel, basis
)
fitter_algo.setOptimizeParameters(True)

# %%
# We run the algorithm and get the metamodel.
fitter_algo.run()
fitter_result = fitter_algo.getResult()
gpr_algo = ot.GaussianProcessRegression(fitter_result)
gpr_algo.run()
gpr_result = gpr_algo.getResult()
gprMetamodel = gpr_result.getMetaModel()

# %%
# Validation of metamodel
# -----------------------
# To validate the metamodel, we create a validation set of size equal to 50 times the input vector dimension to evaluate the functions.
ot.RandomGenerator.SetSeed(1)
experimentTest = ot.LHSExperiment(inputDistribution, 50 * m.dim)
inputTestSet = experimentTest.generate()
outputTestSet = model(inputTestSet)

# %%
# Then, we use the :class:`~openturns.MetaModelValidation` class to validate the metamodel.
metamodelPredictions = gprMetamodel(inputTestSet)
val = ot.MetaModelValidation(outputTestSet, metamodelPredictions)
r2Score = val.computeR2Score()
print("R2=", r2Score)

# %%
# Graphical validation
label = ["Total torque", "Total power", "Solar array area"]
for i in range(3):
    graph = val.drawValidation().getGraph(0, i)
    graph.setLegends([""])
    graph.setLegends(["R2 = %.2f%%" % (100 * r2Score[i]), ""])
    graph.setLegendPosition("upper left")
    graph.setXTitle("Exact function")
    graph.setYTitle("Metamodel prediction")
    graph.setTitle(label[i])
    otv.View(graph)

# %%
otv.View.ShowAll()



#!/usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

# generate reference files
if 0:
    pMin = 0.1
    pMax = 0.9
    nMin = 2
    nMax = 12
    n1 = 5
    nR = 5
    f1 = open("t_DistFunc_binomial1.csv", "w")
    f2 = open("t_DistFunc_binomial2.csv", "w")
    for i1 in range(n1):
        p = pMin + i1 * (pMax - pMin) / (n1 - 1.0)
        for n in range(nMin, nMax + 1):
            for k in range(n + 1):
                f1.write(
                    f"{n}, {p}, {k}, {ot.DistFunc.dBinomial(n, p, k)}, "
                    f"{ot.DistFunc.logdBinomial(n, p, k)}\n"
                )
            for iR in range(nR):
                f2.write(f"{n}, {p}, {iR}, {ot.DistFunc.rBinomial(n, p)}\n")
    f1.close()
    f2.close()
    ot.RandomGenerator.SetSeed(0)

# binomial + logdBinomial
sample1 = ot.Sample.ImportFromCSVFile("t_DistFunc_binomial1.csv", ",")
for x in sample1:
    n, p, k, ref, log_ref = x
    n, k = int(n), int(k)

    val = ot.DistFunc.dBinomial(n, p, k)
    print(f"dbinomial({n}, {p}, {k}) ref={ref} val={val}")
    ott.assert_almost_equal(val, ref)

    log_val = ot.DistFunc.logdBinomial(n, p, k)
    print(f"logdbinomial({n}, {p}, {k}) ref={log_ref} val={log_val}")
    ott.assert_almost_equal(log_val, log_ref)

# rBinomial
sample2 = ot.Sample.ImportFromCSVFile("t_DistFunc_binomial2.csv", ",")
for x in sample2:
    n, p, iR, ref = x
    n, iR = int(n), int(iR)
    val = ot.DistFunc.rBinomial(n, p)
    print(f"rBinomial({n}, {p}) iR={iR} ref={ref} val={val}")
    ott.assert_almost_equal(val, ref)



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv

bipartite = ot.BipartiteGraph([[0, 2], [1, 4], [3, 4, 5]])

graph = bipartite.draw()
graph.setTitle("Bipartite graph")

fig = plt.figure(figsize=(6, 4))
axis = fig.add_subplot(111)
axis.set_xlim(auto=True)
otv.View(graph, figure=fig, axes=[axis], add_legend=True)



"""
Non parametric Adaptive Importance Sampling (NAIS)
==================================================
"""

# %%
#
# The objective is to evaluate a probability from the Non parametric Adaptive Importance Sampling (NAIS) technique.
#
# We consider the four-branch function :math:`g : \mathbb{R}^2 \rightarrow \mathbb{R}` defined by:
#
# .. math::
#   \begin{align*}
#   g(\vect{X}) = \min \begin{pmatrix}5+0.1(x_1-x_2)^2-\frac{(x_1+x_2)}{\sqrt{2}}\\
#   5+0.1(x_1-x_2)^2+\frac{(x_1+x_2)}{\sqrt{2}}\\
#   (x_1-x_2)+ \frac{9}{\sqrt{2}}\\
#   (x_2-x_1)+ \frac{9}{\sqrt{2}}
#   \end{pmatrix}
#   \end{align*}
#
# and the input random vector :math:`\vect{X} = (X_1, X_2)` which follows the standard 2-dimensional Normal distribution:
#
# .. math::
#   \begin{align*}
#   \vect{X} \sim  \mathcal{N}(\mu = [0, 0], \sigma = [1,1], corr = \mat{I}_2)
#   \end{align*}
#
# We want to evaluate the probability:
#
# .. math::
#   \begin{align*}
#   p = \mathbb{P} ( g(\vect{X}) \leq 0 )
#   \end{align*}
#

# %%
# First, import the python modules:

# %%
import openturns as ot
import openturns.viewer as otv
import math

# %%
# Create the probabilistic model :math:`Y = g(\vect{X})`
# ------------------------------------------------------

# %%
# Create the input random vector :math:`\vect{X}`:

# %%
X = ot.RandomVector(ot.Normal(2))

# %%
# Create the function :math:`g` from a :class:`~openturns.PythonFunction`:

# %%


def fourBranch(x):
    x1 = x[0]
    x2 = x[1]

    g1 = 5 + 0.1 * (x1 - x2) ** 2 - (x1 + x2) / math.sqrt(2)
    g2 = 5 + 0.1 * (x1 - x2) ** 2 + (x1 + x2) / math.sqrt(2)
    g3 = (x1 - x2) + 9 / math.sqrt(2)
    g4 = (x2 - x1) + 9 / math.sqrt(2)

    return [min((g1, g2, g3, g4))]


g = ot.PythonFunction(2, 1, fourBranch)

# %%
# Draw the function :math:`g` to help to understand the shape of the limit state function:

# %%
graph = ot.Graph("Four Branch function", "x1", "x2")
graph.setLegendPosition("upper right")
drawfunction = g.draw([-8] * 2, [8] * 2, [100] * 2)
graph.add(drawfunction)
view = otv.View(graph)


# %%
# In order to be able to get the NAIS samples used in the algorithm, it is necessary to transform the :class:`~openturns.PythonFunction` into a :class:`~openturns.MemoizeFunction`:

# %%
g = ot.MemoizeFunction(g)

# %%
# Create the output random vector :math:`Y = g(\vect{X})`:

# %%
Y = ot.CompositeRandomVector(g, X)

# %%
# Create the event :math:`\{ Y = g(\vect{X}) \leq 0 \}`
# -----------------------------------------------------

# %%
threshold = 0.0
myEvent = ot.ThresholdEvent(Y, ot.Less(), threshold)

# %%
# Evaluate the probability with the NAIS technique
# ------------------------------------------------

# %%
quantileLevel = 0.1
algo = ot.NAIS(myEvent, quantileLevel)

# %%
# In order to get all the inputs and outputs that realize the event, you have to mention it now:

# %%
algo.setKeepSample(True)

# %%
# Now you can run the algorithm.

# %%
algo.run()
result = algo.getResult()
proba = result.getProbabilityEstimate()
print("Proba NAIS = ", proba)
print("Current coefficient of variation = ", result.getCoefficientOfVariation())

# %%
# The confidence interval of level :math:`95\%` is:
ci95 = result.getProbabilityDistribution().computeBilateralConfidenceInterval(0.95)
print("Confidence interval (0.95) = ", ci95)

# %%
# You can also get the successive thresholds used by the algorithm:

# %%
levels = algo.getThresholdPerStep()
print("Levels of g = ", levels)

# %%
# Draw the NAIS samples used by the algorithm
# ---------------------------------------------

# %%
# You can get the number :math:`N_s` of steps with:
Ns = algo.getStepsNumber()
print("Number of steps= ", Ns)

# %%
# Get all the inputs where :math:`g` was evaluated at each step
list_subSamples = list()
for step in range(Ns):
    list_subSamples.append(algo.getInputSample(step))

# %%
# The following graph draws each NAIS sample and the frontier :math:`g(x_1, x_2) = l_i` where :math:`l_i` is the threshold at the step :math:`i`:

# %%
graph = ot.Graph()
graph.setAxes(True)
graph.setGrid(True)
graph.setTitle("NAIS sampling: samples")
graph.setXTitle(r"$x_1$")
graph.setYTitle(r"$x_2$")
graph.setLegendPosition("lower left")

# %%
# Add all the NAIS samples:

# %%
for i in range(Ns):
    cloud = ot.Cloud(list_subSamples[i])
    # cloud.setPointStyle("dot")
    graph.add(cloud)
col = graph.getColors()

# %%
# Add the frontiers :math:`g(x_1, x_2) = l_i` where :math:`l_i` is the threshold at the step :math:`i`:

# %%
gIsoLines = g.draw([-5] * 2, [5] * 2, [128] * 2)
dr = gIsoLines.getDrawable(0)
for i, lv in enumerate(levels):
    dr.setLevels([lv])
    dr.setLineStyle("solid")
    dr.setLegend(r"$g(X) = $" + str(round(lv, 2)))
    dr.setLineWidth(3)
    dr.setColor(col[i])
    graph.add(dr)

# %%
_ = otv.View(graph)

# %%
# Draw the frontiers only
# -----------------------
#
# The following graph enables to understand the progression of the algorithm:

# %%
graph = ot.Graph()
graph.setAxes(True)
graph.setGrid(True)
dr = gIsoLines.getDrawable(0)
for i, lv in enumerate(levels):
    dr.setLevels([lv])
    dr.setLineStyle("solid")
    dr.setLegend(r"$g(X) = $" + str(round(lv, 2)))
    dr.setLineWidth(3)
    graph.add(dr)

graph.setColors(col)
graph.setLegendPosition("lower left")
graph.setTitle("NAIS sampling: thresholds")
graph.setXTitle(r"$x_1$")
graph.setYTitle(r"$x_2$")

_ = otv.View(graph)

# %%
# Get all the input and output points that realized the event
# -----------------------------------------------------------
# The following lines are possible only if you have mentioned that you wanted to keep samples with the method *algo.setKeepSample(True)*

# %%
select = ot.NAIS.EVENT1  # points that realize the event
step = Ns - 1  # get the working sample from last iteration
inputEventSample = algo.getInputSample(step, select)
outputEventSample = algo.getOutputSample(step, select)
print("Number of event realizations = ", inputEventSample.getSize())

# %%
# Draw them! They are all in the event space.

# %%
graph = ot.Graph()
graph.setAxes(True)
graph.setGrid(True)
cloud = ot.Cloud(inputEventSample)
cloud.setPointStyle("bullet")
graph.add(cloud)
gIsoLines = g.draw([-5] * 2, [5] * 2, [1000] * 2)
dr = gIsoLines.getDrawable(0)
dr.setLevels([0.0])
dr.setColor("red")
graph.add(dr)
_ = otv.View(graph)

otv.View.ShowAll()



"""
Select fitted distributions
===========================
"""

# %%
# In this example help to make a choice between several distributions fitted to a sample.
#
# Several methods can be used:
#
# - the ranking by the Kolmogorov p-values (for continuous distributions),
# - the ranking by the ChiSquared p-values (for discrete distributions),
# - the ranking by BIC values.

# %%
import openturns as ot
import openturns.viewer as otv


# %%
# Create a sample from a continuous distribution
distribution = ot.Beta(2.0, 2.0, 0.0, 1.0)
sample = distribution.getSample(1000)
graph = ot.FiniteDiscreteDistribution(sample).drawCDF()
view = otv.View(graph)

# %%
# **1. Specify the model only**

# %%
# Create the list of distribution estimators
factories = [ot.BetaFactory(), ot.TriangularFactory()]

# %%
# Rank the continuous models by the Lilliefors p-values:
estimated_distribution, test_result = ot.FittingTest.BestModelLilliefors(
    sample, factories
)
test_result

# %%
# Rank the continuous models wrt the BIC criteria (no test result):
ot.FittingTest.BestModelBIC(sample, factories)

# %%
# Rank the continuous models wrt the AIC criteria (no test result)
ot.FittingTest.BestModelAIC(sample, factories)

# %%
# Rank the continuous models wrt the AICc criteria (no test result):
ot.FittingTest.BestModelAICC(sample, factories)

# %%
# **2. Specify the model and its parameters**

# %%
# Create a collection of the distributions to be tested
distributions = [ot.Beta(2.0, 2.0, 0.0, 1.0), ot.Triangular(0.0, 0.5, 1.0)]

# %%
# Rank the continuous models by the Kolmogorov p-values:
estimated_distribution, test_result = ot.FittingTest.BestModelKolmogorov(
    sample, distributions
)
test_result

# %%
# Rank the continuous models wrt the BIC criteria:
ot.FittingTest.BestModelBIC(sample, distributions)

# %%
# Rank the continuous models wrt the AIC criteria:
ot.FittingTest.BestModelAIC(sample, distributions)

# %%
# Rank the continuous models wrt the AICc criteria:
ot.FittingTest.BestModelAICC(sample, distributions)

# %%
# **Discrete distributions**

# %%
# Create a sample from a discrete distribution
distribution = ot.Poisson(2.0)
sample = distribution.getSample(1000)
graph = ot.FiniteDiscreteDistribution(sample).drawCDF()
view = otv.View(graph)

# %%
# Create the list of distribution estimators
distributions = [ot.Poisson(2.0), ot.Geometric(0.1)]

# %%
# Rank the discrete models with respect to the ChiSquared p-values:
ot.ResourceMap.SetAsBool("FittingTest-ChiSquaredCheckSample", False)
estimated_distribution, test_result = ot.FittingTest.BestModelChiSquared(
    sample, distributions
)
test_result

# %%
# Rank the discrete models with respect to the BIC criteria:
ot.FittingTest.BestModelBIC(sample, distributions)
otv.View.ShowAll()



import sys
from os.path import basename
import openturns as ot


from io import StringIO
from docutils.parsers.rst import Directive
from docutils import nodes


class execforresourcemap_directive(Directive):
    """Execute the specified python code and insert the output into the document"""

    has_content = True

    def run(self):
        oldStdout, sys.stdout = sys.stdout, StringIO()
        source = self.state_machine.input_lines.source(
            self.lineno - self.state_machine.input_offset - 1
        )
        try:
            table = nodes.table()

            tgroup = nodes.tgroup(cols=3)
            table += tgroup
            tgroup += nodes.colspec(colwidth=25, classes=["key"])
            tgroup += nodes.colspec(colwidth=8, classes=["value"])
            tgroup += nodes.colspec(colwidth=8, classes=["type"])
            thead = nodes.thead()
            tgroup += thead

            # Add headers
            row = nodes.row()
            thead += row

            entry = nodes.entry()
            row += entry
            node = nodes.paragraph(text="Key")
            entry += node

            entry = nodes.entry()
            row += entry
            node = nodes.paragraph(text="Value")
            entry += node

            entry = nodes.entry()
            row += entry
            node = nodes.paragraph(text="Type")
            entry += node

            # Add body
            tbody = nodes.tbody()
            tgroup += tbody
            row = nodes.row()
            tbody += row

            # reset state in case altered by examples
            ot.ResourceMap.Reset()

            for key in ot.ResourceMap.GetKeys():
                row = nodes.row()
                tbody += row

                entry = nodes.entry()
                row += entry

                node = nodes.paragraph(text=key)
                entry += node

                entry = nodes.entry()
                row += entry

                value = ot.ResourceMap.Get(key)
                if not len(value):
                    value = " ".__repr__()
                if "\t" in value:
                    value = value.replace("\t", "\\t")
                node = nodes.paragraph(text=value)
                entry += node

                entry = nodes.entry()
                row += entry

                keyType = ot.ResourceMap.GetType(key)
                node = nodes.paragraph(text=keyType)
                entry += node

            return [table]
        except Exception:
            return [
                nodes.error(
                    None,
                    nodes.paragraph(
                        text="Unable to execute python code at %s:%d:"
                        % (basename(source), self.lineno)
                    ),
                    nodes.paragraph(text=str(sys.exc_info()[1])),
                )
            ]
        finally:
            sys.stdout = oldStdout


def setup(app):
    setup.app = app
    setup.config = app.config
    setup.confdir = app.confdir
    app.add_directive("execforresourcemap", execforresourcemap_directive)

    metadata = {"version": "0", "parallel_read_safe": True}
    return metadata



#! /usr/bin/env python

import openturns as ot
import math as m

ot.ResourceMap.Set("SymbolicParser-Backend", "MuParser")

parsers = ot.SymbolicFunction.GetValidParsers()
assert "MuParser" in parsers, "MuParser not found"

elementaryFunctions = ["sin", "cos", "tan"]
elementaryFunctions.append("asin")
elementaryFunctions.append("acos")
elementaryFunctions.append("atan")
elementaryFunctions.append("sinh")
elementaryFunctions.append("cosh")
elementaryFunctions.append("tanh")
elementaryFunctions.append("asinh")
elementaryFunctions.append("acosh")
elementaryFunctions.append("atanh")
elementaryFunctions.append("log2")
elementaryFunctions.append("log10")
elementaryFunctions.append("log")
elementaryFunctions.append("ln")
elementaryFunctions.append("lngamma")
elementaryFunctions.append("gamma")
elementaryFunctions.append("exp")
elementaryFunctions.append("erf")
elementaryFunctions.append("erfc")
elementaryFunctions.append("sqrt")
elementaryFunctions.append("cbrt")
elementaryFunctions.append("besselJ0")
elementaryFunctions.append("besselJ1")
elementaryFunctions.append("besselY0")
elementaryFunctions.append("besselY1")
elementaryFunctions.append("sign")
elementaryFunctions.append("rint")
elementaryFunctions.append("abs")
elementaryFunctions.append("min")
elementaryFunctions.append("max")
elementaryFunctions.append("sum")
elementaryFunctions.append("avg")
elementaryFunctions.append("floor")
elementaryFunctions.append("ceil")
elementaryFunctions.append("trunc")
elementaryFunctions.append("round")


# Check the creation of the elementary functions
for func in elementaryFunctions:
    x = [0.4]
    # acosh only defined for 1 <= x <= pi
    if func == "acosh":
        x[0] = 1.4

    f = ot.SymbolicFunction(["x"], ["2.0*" + func + "(x)"])
    print("f=", f)
    print("f(", x[0], ")=%.4e" % f(x)[0])
    try:
        df = f.gradient(x)[0, 0]
    except Exception:
        pass
    else:
        f.setGradient(
            ot.CenteredFiniteDifferenceGradient(
                ot.ResourceMap.GetAsScalar(
                    "CenteredFiniteDifferenceGradient-DefaultEpsilon"
                ),
                f.getEvaluation(),
            )
        )
        df2 = f.gradient(x)[0, 0]
        print("df(", x[0], ")=%.4e" % df, "df (FD)=%.4e" % df2)
        if abs(df) > 1e-5:
            err_g = abs(df2 / df - 1.0)
        else:
            err_g = abs(df - df2)
        if err_g > 1e-5:
            print("GRADIENT ERROR! check " + func + " gradient, err=%.12g" % err_g)
    try:
        d2f = f.hessian(x)[0, 0, 0]
    except Exception:
        pass
    else:
        f.setHessian(
            ot.CenteredFiniteDifferenceHessian(
                ot.ResourceMap.GetAsScalar(
                    "CenteredFiniteDifferenceHessian-DefaultEpsilon"
                ),
                f.getEvaluation(),
            )
        )
        d2f2 = f.hessian(x)[0, 0, 0]
        print("d2f(", x[0], ")=%.4e" % d2f, "d2f (FD)=%.4e" % d2f2)
        if abs(d2f) > 1e-5:
            err_h = abs(d2f2 / d2f - 1.0)
        else:
            err_h = abs(d2f - d2f2)
        if err_h > 1e-4:
            print("HESSIAN ERROR! check " + func + " hessian, err=%.12g" % err_h)

nmf = ot.SymbolicFunction(["x0", "x1"], ["x0+x1", "x0-x1"])
marginal0 = nmf.getMarginal(0)
marginal1 = nmf.getMarginal(1)
print("marginal 0=", marginal0)
print("marginal 1=", marginal1)

# test sample as input of a function
formula = ["sin(xi1) + 7. * (sin(xi2)) ^ 2 + 0.1 * xi3^4 * sin(xi1)"]
model = ot.SymbolicFunction(["xi1", "xi2", "xi3"], formula)

# Create an input distribution to calculate reference values
distribution = ot.JointDistribution([ot.Uniform(-m.pi, m.pi)] * 3)
inSample = distribution.getSample(100)
resultSample = model(inSample)
refResultValues = [
    m.sin(inSample[i][0])
    + 7.0 * (m.sin(inSample[i][1])) ** 2
    + 0.1 * (inSample[i][2]) ** 4 * m.sin(inSample[i][0])
    for i in range(100)
]

print("First reference value : %.4e" % refResultValues[0])
print("First result calculated : %.4e" % resultSample[0][0])

model = ot.SymbolicFunction(["x"], ["pi_", "e_"])
print("Constants:", model([0]))

empty = model.getMarginal([])
x = [42.0]
y = empty(x)
print("empty eval", y)

constants = ot.SymbolicFunction.GetValidConstants()
assert len(constants) > 0, "empty"
funcs = ot.SymbolicFunction.GetValidFunctions()
assert len(funcs) > 0, "empty"
ops = ot.SymbolicFunction.GetValidOperators()
assert len(ops) > 0, "empty"
print("OK")

# str ctor
model = ot.SymbolicFunction("x", "3*x")
assert model(x)[0] == 3.0 * x[0], "str ctor eval"

# Check constants
f = ot.SymbolicFunction("x", "e_")
print(f, ", e_=", f([0.0]))
f = ot.SymbolicFunction("x", "pi_")
print(f, ", pi_=", f([0.0]))
# Check unary operators
f = ot.SymbolicFunction("x", "-x")
print(f, ", f([1])=", f([1.0]))
f = ot.SymbolicFunction("x", "(x=2.0)*x")
print(f, ", f([1])=", f([1.0]))
f = ot.SymbolicFunction("x", "!x")
print(f, ", f([0])=", f([0.0]))
print(f, ", f([1])=", f([1.0]))
print(f, ", f([1.5])=", f([1.5]))
print(f, ", f([-0.5])=", f([-0.5]))
# Check binary operators
f = ot.SymbolicFunction(["x", "y"], ["x && y"])
print(f, ", f([0, 0])=", f([0.0, 0.0]))
print(f, ", f([1, 0])=", f([1.0, 0.0]))
print(f, ", f([1, 1])=", f([1.0, 1.0]))
print(f, ", f([0, 1])=", f([0.0, 1.0]))
print(f, ", f([1.5, -0.5])=", f([1.5, -0.5]))
f = ot.SymbolicFunction(["x", "y"], ["x || y"])
print(f, ", f([0, 0])=", f([0.0, 0.0]))
print(f, ", f([1, 0])=", f([1.0, 0.0]))
print(f, ", f([1, 1])=", f([1.0, 1.0]))
print(f, ", f([0, 1])=", f([0.0, 1.0]))
print(f, ", f([1.5, -0.5])=", f([1.5, -0.5]))
f = ot.SymbolicFunction(["x", "y"], ["x ^^ y"])
print(f, ", f([0, 0])=", f([0.0, 0.0]))
print(f, ", f([1, 0])=", f([1.0, 0.0]))
print(f, ", f([1, 1])=", f([1.0, 1.0]))
print(f, ", f([0, 1])=", f([0.0, 1.0]))
print(f, ", f([1.5, -0.5])=", f([1.5, -0.5]))
f = ot.SymbolicFunction(["x", "y"], ["x <= y"])
print(f, ", f([1, 2])=", f([1.0, 2.0]))
print(f, ", f([1, 1])=", f([1.0, 1.0]))
print(f, ", f([2, 1])=", f([2.0, 1.0]))
f = ot.SymbolicFunction(["x", "y"], ["x >= y"])
print(f, ", f([1, 2])=", f([1.0, 2.0]))
print(f, ", f([1, 1])=", f([1.0, 1.0]))
print(f, ", f([2, 1])=", f([2.0, 1.0]))
f = ot.SymbolicFunction(["x", "y"], ["x != y"])
print(f, ", f([1, 2])=", f([1.0, 2.0]))
print(f, ", f([1, 1])=", f([1.0, 1.0]))
print(f, ", f([2, 1])=", f([2.0, 1.0]))
f = ot.SymbolicFunction(["x", "y"], ["x == y"])
print(f, ", f([1, 2])=", f([1.0, 2.0]))
print(f, ", f([1, 1])=", f([1.0, 1.0]))
print(f, ", f([2, 1])=", f([2.0, 1.0]))
f = ot.SymbolicFunction(["x", "y"], ["x > y"])
print(f, ", f([1, 2])=", f([1.0, 2.0]))
print(f, ", f([1, 1])=", f([1.0, 1.0]))
print(f, ", f([2, 1])=", f([2.0, 1.0]))
f = ot.SymbolicFunction(["x", "y"], ["x < y"])
print(f, ", f([1, 2])=", f([1.0, 2.0]))
print(f, ", f([1, 1])=", f([1.0, 1.0]))
print(f, ", f([2, 1])=", f([2.0, 1.0]))
f = ot.SymbolicFunction(["x", "y"], ["x + y"])
print(f, ", f([1, 2])=", f([1.0, 2.0]))
f = ot.SymbolicFunction(["x", "y"], ["x - y"])
print(f, ", f([1, 2])=", f([1.0, 2.0]))
f = ot.SymbolicFunction(["x", "y"], ["x * y"])
print(f, ", f([2, 3])=", f([2.0, 3.0]))
f = ot.SymbolicFunction(["x", "y"], ["x / y"])
print(f, ", f([2, 3])=", f([2.0, 3.0]))
f = ot.SymbolicFunction(["x", "y"], ["x ^ y"])
print(f, ", f([2, 3])=", f([2.0, 3.0]))
# Check for exceptional output
f = ot.SymbolicFunction("x", "sqrt(x)")
try:
    print(f, ", f([-1])=", f([-1.0]))
except Exception:
    print(f, ", f([-1]) not defined")

f = ot.SymbolicFunction("x", "sqrt(x)")
ev = f.getEvaluation()
# triggers copyOnWrite, ev is no longer the Evaluation of f
ev.setCheckOutput(False)
print(f, ", f([-1]) is normal?", m.isfinite(ev([-1.0])[0]))



#! /usr/bin/env python

import openturns as ot
import math as m

ot.TESTPREAMBLE()


# Instantiate one distribution object
dim = 1
meanPoint = ot.Point(dim, 1.0)
meanPoint[0] = 0.5
sigma = ot.Point(dim, 1.0)
sigma[0] = 2.0
R = ot.CorrelationMatrix(dim)

distribution1 = ot.Normal(meanPoint, sigma, R)

# Instantiate another distribution object
meanPoint[0] = -1.5
sigma[0] = 4.0

distribution2 = ot.Normal(meanPoint, sigma, R)

# Test for sampling
size = 2000
nBars = 20
sample1 = distribution1.getSample(size)
sample2 = distribution2.getSample(size)

# Construct histograms
epsilon = 0.1
min1 = sample1.getMin()[0]
max1 = sample1.getMax()[0] + epsilon
min2 = sample2.getMin()[0]
max2 = sample2.getMax()[0] + epsilon
tmp = ot.Point(2)
tmp[0] = (max1 - min1) / nBars
data1 = ot.Sample(nBars, tmp)
tmp[0] = (max2 - min2) / nBars
data2 = ot.Sample(nBars, tmp)

for i in range(size):
    index = int(m.floor((sample1[i, 0] - min1) / (max1 - min1) * nBars))
    data1[index, 1] += 1
    index = int(m.floor((sample2[i, 0] - min2) / (max2 - min2) * nBars))
    data2[index, 1] += 1

# Create an empty graph
myGraph = ot.Graph("Some barplots", "y", "frequency")
myGraph.setLegendPosition("upper left")

# Create the first barplot
myBarPlot1 = ot.BarPlot(data1, min1, "blue", "shaded", "dashed", "histogram1")

# Then, draw it
myGraph.add(myBarPlot1)

# Check that the correct files have been generated by computing their
# checksum

# Create the second barplot
myBarPlot2 = ot.BarPlot(data2, min2, "red", "solid", "solid", "histogram2")

# Add it to the graph and draw everything
myGraph.add(myBarPlot2)



"""
Gaussian Process Regression : generate trajectories from the metamodel
======================================================================
"""

# %%
# The main goal of this example is to show how to simulate new trajectories from a Gaussian Process Regression metamodel(refer to :ref:`gaussian_process_regression`).

# %%
# Introduction
# ------------
#
# We consider the sine function:
#
# .. math::
#    \model(x) = \sin(x)
#
#
# for any :math:`x\in[0,12]`.
#
# We want to create a metamodel of this function. This is why we create a sample of :math:`n` observations of the function:
#
# .. math::
#    y_i = \model(x_i)
#
#
# for :math:`i=1,...,7`, where :math:`x_i` is the i-th input and :math:`y_i` is the corresponding output.
#
# We consider the seven following inputs :
#
# ============ === === === === ===== ==== ======
#  :math:`i`    1   2   3   4    5     6    7
# ============ === === === === ===== ==== ======
#  :math:`x_i`  1   3   4   6   7.9   11   11.5
# ============ === === === === ===== ==== ======
#
# We are going to consider a Gaussian Process Regression metamodel with:
#
# * a constant trend,
# * a Matern covariance model.
#
# In the :doc:`/auto_surrogate_modeling/gaussian_process_regression/plot_gpr_1d` example, we detail the estimation of this metamodel.
# Refer to it for further details: we only focus here on the simulation of new trajectories.
#
import openturns as ot
import openturns.viewer as otv
import openturns.experimental as otexp


# %%
# Creation of the metamodel
# -------------------------
#
# We define the function `g`, the training sample `(x_train, y_train)` and the test sample `(x_test, y_tst)`.

# %%
g = ot.SymbolicFunction(["x"], ["sin(x)"])

# %%
x_train = ot.Sample([[x] for x in [1.0, 3.0, 4.0, 6.0, 7.9, 11.0, 11.5]])
y_train = g(x_train)
n_train = x_train.getSize()
n_train

# %%
xmin = 0.0
xmax = 12.0
n_test = 101
step = (xmax - xmin) / (n_test - 1)
myRegularGrid = ot.RegularGrid(xmin, step, n_test)
x_test = myRegularGrid.getVertices()
y_test = g(x_test)


# %%
# In order to observe the function and the location of the points in the input design of experiments, we define the following function which plots the data.


# %%


def plot_1d_data(x_data, y_data, type="Curve", legend=None, color=None, linestyle=None):
    """Plot the data (x_data,y_data) as a Cloud/Curve"""
    if type == "Curve":
        graphF = ot.Curve(x_data, y_data)
    else:
        graphF = ot.Cloud(x_data, y_data)
    if legend is not None:
        graphF.setLegend(legend)
    if color is not None:
        graphF.setColor(color)
    if linestyle is not None:
        graphF.setLineStyle(linestyle)
    return graphF


# %%
# Here, we draw the model and the train sample.
graph = ot.Graph("Model and Train sample", "X", "Y")
graph.add(
    plot_1d_data(
        x_test, y_test, type="Curve", legend="model", color="black", linestyle="dashed"
    )
)
graph.add(
    plot_1d_data(x_train, y_train, type="Cloud", legend="train sample", color="red")
)
graph.setLegendPosition("upper right")
view = otv.View(graph)

# %%
# We use the :class:`~openturns.ConstantBasisFactory` class to define the trend and the :class:`~openturns.MaternModel` class to define the covariance model.
# This Matérn model is based on the regularity parameter :math:`\nu=3/2`.

# %%
dimension = 1
basis = ot.ConstantBasisFactory(dimension).build()
covarianceModel = ot.MaternModel([1.0] * dimension, 1.5)

# %%
# We estimate the Gaussian process :math:`Y` with the class :class:`~openturns.GaussianProcessFitter`.
fitter_algo = ot.GaussianProcessFitter(x_train, y_train, covarianceModel, basis)
fitter_algo.run()
fitter_result = fitter_algo.getResult()
print(fitter_result)

# %%
# We observe that the `scale` and `amplitude` hyper-parameters have been optimized by the
# :meth:`~openturns.GaussianProcessFitter.run` method, while the :math:`\nu`
# parameter has remained unchanged, as expected.
#
# Then, we condition the gaussian process to make it interpolate the data set using the class
# :class:`~openturns.GaussianProcessRegression`.
gpr_algo = ot.GaussianProcessRegression(fitter_result)
gpr_algo.run()
gpr_result = gpr_algo.getResult()
print(gpr_result)

# %%
# We get the metamodel and the predictions on the test sample.
gpr_metamodel = gpr_result.getMetaModel()
y_test_MM = gpr_metamodel(x_test)

# %%
# The following function plots the Gaussian Process Regression predictions on the test sample.
graph = ot.Graph("Gaussian process regression metamodel", "X", "Y")
graph.add(
    plot_1d_data(
        x_test, y_test, type="Curve", legend="model", color="black", linestyle="dashed"
    )
)
graph.add(
    plot_1d_data(x_train, y_train, type="Cloud", legend="train sample", color="red")
)
graph.add(
    plot_1d_data(
        x_test, y_test_MM, type="Curve", legend="GPR", color="blue", linestyle="solid"
    )
)
graph.setLegendPosition("upper right")
view = otv.View(graph)

# %%
# Simulate new trajectories
# -------------------------
#
# In order to generate new trajectories of the conditioned Gaussian process, we use the class
# :class:`~openturns.experimental.ConditionedGaussianProcess`, which provides a :class:`~openturns.Process`.
# It is created from the result of the Gaussian Process Regression algorithm.
process = otexp.ConditionedGaussianProcess(gpr_result, myRegularGrid)

# %%
# The method :meth:`~openturns.Process.getSample` method returns a :class:`~openturns.ProcessSample`.

# %%
# sphinx_gallery_thumbnail_number = 3
trajectories = process.getSample(10)
type(trajectories)
graph = trajectories.drawMarginal()
graph.add(
    plot_1d_data(
        x_test,
        y_test,
        type="Curve",
        legend="model",
        color="black",
        linestyle="dashed",
    )
)
graph.add(
    plot_1d_data(x_train, y_train, type="Cloud", legend="train sample", color="red")
)
graph.setXTitle("X")
graph.setYTitle("Y")
graph.setLegendPosition("upper right")
graph.setTitle("10 simulated trajectories")
view = otv.View(graph)

# %%
# Display all figures.
otv.View.ShowAll()



"""
Compute confidence intervals of a regression model from data
============================================================
"""

# %%
#
# Introduction
# ------------
#
# In this example, we compute the pointwise confidence interval of the
# estimator of the conditional expectation given an input.
# More precisely, we compute confidence intervals of the output of
# a regression model.
# The linear regression model is an order 1 multivariate polynomial.
# This model is built from a data set.
# In this advanced example, we use the :class:`~openturns.DesignProxy`
# and :class:`~openturns.QRMethod` low level classes.
# Another example of this method is presented in
# :doc:`/auto_numerical_methods/general_methods/plot_regression_sinus`.

# %%

import openturns as ot
import openturns.viewer as otv
import numpy as np
from openturns.usecases import linthurst


# %%
palette = ot.Drawable.BuildTableauPalette(5)

# %%
#
# Get the data
# ------------
#
# We consider the so-called :ref:`Linthurst<use-case-linthurst>` data set, which contains measures of
# aerial biomass (BIO) as well as 5 five physico-chemical properties of
# the soil: salinity (SAL), pH, K, Na, and Zn.
# The data set is taken from [rawlings2001]_ table 5.1 page 63.


# %%
# We define the data.
#

# %%
ds = linthurst.Linthurst()


# %%
# Get the input and output samples.

# %%
dimension = ds.data.getDimension() - 1
print("dimension = ", dimension)
sampleSize = ds.data.getSize()
print("sampleSize = ", sampleSize)
inputSample = ds.data[:, 1 : dimension + 1]
print("Input :")
print(inputSample[:5])
outputSample = ds.data[:, 0]
print("Output :")
print(outputSample[:5])

inputDescription = inputSample.getDescription()
outputDescription = outputSample.getDescription()


# %%
# We want to create a linear regression model.
# To do this, we need to create a functional basis.
# We make the choice of using only degree 1 polynomials
# for each marginal input variable.

functionCollection = []
basisFunction = ot.SymbolicFunction(inputDescription, ["1.0"])
functionCollection.append(basisFunction)
for i in range(dimension):
    basisFunction = ot.SymbolicFunction(inputDescription, [inputDescription[i]])
    functionCollection.append(basisFunction)
basis = ot.Basis(functionCollection)


# %%
# We can then print the basis.

basisSize = basis.getSize()
print("Basis size = ", basisSize)
for i in range(basisSize):
    basisFunction = basis.build(i)
    print("Function #", i, basisFunction)


# %%
# Create the least squares model
# ------------------------------
#
# To create the least squares model, we use the :class:`~openturns.DesignProxy` class,
# which defines the design matrix of the linear regression model.
# Then we solve the problem using QR decomposition.

designProxy = ot.DesignProxy(inputSample, basis)
indices = range(basisSize)  # We consider all the functions in the basis
lsqMethod = ot.QRMethod(designProxy, indices)
betaHat = lsqMethod.solve(outputSample.asPoint())
print(betaHat)

# %%
# Based on the solution of the linear least squares problem, we
# can create the metamodel and evaluate the residuals.

metamodel = ot.LinearCombinationFunction(basis, betaHat)

# %%
# Compute residuals, variance and design matrix
# ---------------------------------------------
# We need to estimate the variance of the residuals.
# To do this we evaluate the predictions of the regression model
# on the training sample and compute the residuals.
yHat = metamodel(inputSample).asPoint()
residuals = yHat - outputSample.asPoint()

# %%
# In order to compute confidence intervals of the mean, we need
# to estimate the sample standard deviation.

sigma2Hat = residuals.normSquare() / (sampleSize - basisSize)
print("sigma2Hat", sigma2Hat)
sigmaHat = np.sqrt(sigma2Hat)
print("sigmaHat = ", sigmaHat)

# %%
# We evaluate the design matrix and store it as a :class:`~openturns.Sample`.

designMatrix = lsqMethod.computeWeightedDesign()
designSample = ot.Sample(np.array(ot.Matrix(designMatrix)))


# %%
# Compute confidence intervals
# ----------------------------
#
# The next function will help to compute confidence intervals.
# It is based on regression analysis.


def computeRegressionConfidenceInterval(
    lsqMethod,
    betaHat,
    sigma2Hat,
    designSample,
    alpha=0.95,
    mean=True,
    epsilonSigma=1.0e-5,
):
    """
    Compute a confidence interval for the estimate of the mean.

    Evaluates this confidence interval at points in the design matrix.

    This code is based on (Rawlings, Pantula & David, 1998)
    eq. 3.51 and 3.52 page 90.

    Parameters
    ----------
    lsqMethod: ot.LeastSquaresMethod
        The linear least squares method (e.g. QR or Cholesky).
    betaHat : ot.Point(parameterDimension)
        The solution of the least squares problem.
    sigma2Hat : float > 0.0
        The estimate of the variance.
    designSample : ot.Sample(size, parameterDimension)
        The design matrix of the linear model.
        This is the value of the functional basis depending on the
        input sample.
        Each row represents the input where the confidence interval
        is to be computed.
    alpha : float, in [0, 1]
        The width of the confidence interval.
    mean : bool
        If True, then computes the confidence interval of the mean.
        This interval contains yTrue = E[y|x] with probability alpha.
        Otherwise, computes a confidence interval of the prediction at point x.
        This interval contains y|x with probability alpha.
    epsilonSigma : float > 0.0
        A relatively small number. The minimal value of variance, which
        avoids a singular Normal distribution.

    Reference
    ---------
    - O. Rawlings John, G. Pantula Sastry, and A. Dickey David.
      Applied regression analysis—a research tool. Springer New York, 1998.

    Returns
    -------
    confidenceBounds : ot.Sample(size, 2)
        The first column contains the lower bound.
        The second column contains the upper bound.
    """

    inverseGram = lsqMethod.getGramInverse()
    sampleSize = designSample.getSize()
    confidenceBounds = ot.Sample(sampleSize, 2)
    for i in range(sampleSize):
        x0 = designSample[i, :]
        meanYHat = x0.dot(betaHat)
        sigma2YHat = x0.dot(inverseGram * x0) * sigma2Hat
        if not mean:
            sigma2YHat += sigma2Hat
        sigmaYHat = np.sqrt(sigma2YHat)
        sigmaYHat = max(epsilonSigma, sigmaYHat)  # Prevents a zero s.e.
        distribution = ot.Normal(meanYHat, sigmaYHat)
        interval = distribution.computeBilateralConfidenceInterval(alpha)
        lower = interval.getLowerBound()
        upper = interval.getUpperBound()
        confidenceBounds[i, 0] = lower[0]
        confidenceBounds[i, 1] = upper[0]
    return confidenceBounds


# %%
# We evaluate the value of the basis functions on the test sample.
# This sample is used to compute the confidence interval.

alpha = 0.95
confidenceIntervalMean = computeRegressionConfidenceInterval(
    lsqMethod, betaHat, sigma2Hat, designSample, alpha=alpha
)


# %%
# For a given observation, we can print the input, the observed
# output, the predicted output and the confidence interval of
# the conditional expectation.

i = 5  # The index of the observation
print("x = ", inputSample[i])
print("Y observation = ", outputSample[i])
print("Y prediction = ", yHat[i])
print("Confidence interval of the mean = ", confidenceIntervalMean[i])


# %%
# In order to see how the model fits to the data, we can
# create the validation plot.
# Each vertical bar represents the 95% confidence interval
# of the estimate of the conditional expectation of the linear regression model.

metamodelPredictions = metamodel(inputSample)
validation = ot.MetaModelValidation(outputSample, metamodelPredictions)
graph = validation.drawValidation().getGraph(0, 0)
q2Score = validation.computeR2Score()[0]
graph.setTitle("Q2 = %.2f%%" % (100.0 * q2Score))
graph.setXTitle("Observations")
graph.setYTitle("Metamodel")
for i in range(sampleSize):
    curve = ot.Curve([outputSample[i, 0]] * 2, confidenceIntervalMean[i])
    graph.add(curve)
view = otv.View(graph)

# %%
# We see that the linear regression model is not a very accurate
# metamodel, as can be seen from the relatively low R2 score.
# The metamodel predictions are not very close to observations,
# which is why the points are not close to the diagonal of the plot.
# Hence, the confidence intervals do not cross the diagonal very often.
# In this case, we may want to create a more accurate metamodel.

# %%
# Display all figures
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot

dimension = 2

# Size of sample
size = 20

# Factory: lhs generates
lhsDesign = ot.LHSExperiment(
    ot.JointDistribution([ot.Uniform(20.0, 30.0)] * dimension), size
)
lhsDesign.setRandomShift(False)  # centered
lhsDesign.setAlwaysShuffle(True)  # randomized


# Number of designs to generate
N = 10

# For each design, evaluation of some criteria
fp = open("space_filling_criteria_nonunit.val", "w")
for i in range(N):
    design = lhsDesign.generate()
    filename = "design_dim_%d_size_%d_%d_nonunit.csv" % (dimension, size, i + 1)
    design.exportToCSVFile(filename, ";")
    c2 = ot.SpaceFillingC2().evaluate(design)
    phip = ot.SpaceFillingPhiP().evaluate(design)
    mindist = ot.SpaceFillingMinDist().evaluate(design)
    fp.write("design=%s\n" % filename)
    fp.write("c2=%1.13e, phip=%1.13e, mindist=%1.13e\n" % (c2, phip, mindist))

# New test with biggest dimension
dimension = 50

# Size of sample
size = 100
lhsDesign = ot.LHSExperiment(
    ot.JointDistribution([ot.Uniform(20.0, 30.0)] * dimension), size
)
lhsDesign.setRandomShift(False)  # centered
lhsDesign.setAlwaysShuffle(True)  # randomized

# Number of designs to generate
N = 5

# For each design, evaluation of some criteria
for i in range(N):
    design = lhsDesign.generate()
    filename = "design_dim_%d_size_%d_%d_nonunit.csv" % (dimension, size, i + 1)
    design.exportToCSVFile(filename, ";")
    c2 = ot.SpaceFillingC2().evaluate(design)
    phip = ot.SpaceFillingPhiP().evaluate(design)
    mindist = ot.SpaceFillingMinDist().evaluate(design)
    fp.write("design=%s\n" % filename)
    fp.write("c2=%1.13e, phip=%1.13e, mindist=%1.13e\n" % (c2, phip, mindist))
fp.close()



#! /usr/bin/env python

import openturns as ot
from openturns.testing import assert_almost_equal
from openturns.usecases import logistic_model


ot.TESTPREAMBLE()
ot.PlatformInfo.SetNumericalPrecision(5)

"""
Test the import of the LogisticModel data class.
"""
lm = logistic_model.LogisticModel()

# test parameters
assert_almost_equal(lm.y0, 3.9e6, 1e-12)
assert_almost_equal(lm.a, 0.03134, 1e-12)
assert_almost_equal(lm.b, 1.5887e-10, 1e-12)

# test marginals means
assert_almost_equal(lm.distY0.getMean()[0], 3.9e6, 1e-12)
assert_almost_equal(lm.distA.getMean()[0], 0.03134, 1e-12)
assert_almost_equal(lm.distB.getMean()[0], 1.5887e-10, 1e-12)



import openturns as ot
import openturns.viewer as otv


sampleX = ot.Normal(2).getSample(100)

f = ot.SymbolicFunction(["x1", "x2"], ["x1 + x2", "2 * x1 - x2"])

sampleY = f(sampleX)

graph = ot.VisualTest.DrawPairsXY(sampleX, sampleY)

otv.View(graph, figure_kw={"figsize": (6.0, 6.0)})



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


sample1 = ot.Sample(5, ot.Point(1, 2.0))
sample2 = ot.Sample(sample1)
sample2[0] = ot.Point(1, 3.0)
print("sample1=", repr(sample1))
print("sample2=", repr(sample2))
sample1[1] = ot.Point(1, -5.0)
print("sample1=", repr(sample1))
print("sample2=", repr(sample2))
sample1[3, 0] = 6.0
print("sample1=", repr(sample1))
print("sample1[2][0]=", sample1[2][0])
print("sample1[2, 0]=", sample1[2, 0])
try:
    sample1[1, 2] = -4.0
except Exception:
    print("set item failed on purpose")



#! /usr/bin/env python

import openturns as ot
from openturns.testing import assert_almost_equal
from openturns.usecases import flood_model


ot.TESTPREAMBLE()
ot.PlatformInfo.SetNumericalPrecision(5)

print("Test the FloodModel data class.")
fm = flood_model.FloodModel()

# test parameter
assert fm.dim == 8

# test marginals means
assert_almost_equal(fm.Q.getMean()[0], 1338.1299666394023, 1e-12)
assert_almost_equal(fm.Ks.getMean()[0], 30.001003758482675, 1e-12)
assert_almost_equal(fm.Zv.getMean()[0], 50.0, 1e-12)
assert_almost_equal(fm.Zm.getMean()[0], 55.0, 1e-12)
assert_almost_equal(fm.B.getMean()[0], 300.0, 1e-12)
assert_almost_equal(fm.L.getMean()[0], 5000.0, 1e-12)
assert_almost_equal(fm.Zb.getMean()[0], 55.5, 1e-12)
assert_almost_equal(fm.Hd.getMean()[0], 3.0, 1e-12)

meanInput = fm.distribution.getMean()
print("meanInput = ", meanInput)
meanOutput = fm.model(meanInput)
print("meanOutput = ", meanOutput)
assert_almost_equal(meanOutput, [2.5313, -5.96869, 1.0362], 1.0e-4)

#
print("Test the simplified height model")
heightInputDistribution, heightModel = fm.getHeightModel()
meanInput = heightInputDistribution.getMean()
print("meanInput = ", meanInput)
meanOutput = heightModel(meanInput)
print("meanOutput = ", meanOutput)
assert_almost_equal(meanOutput, [2.5313])

print("Test the high Hd scenario")
fm = flood_model.FloodModel(distributionHdLow=False)
assert_almost_equal(fm.Hd.getMean()[0], 8.0, 1e-12)



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

# Latent variable model for 4 categorical levels and
# a 3-dimensional latent space
k = ot.LatentVariableModel(4, 3)
k.setLatentVariables([0.1, 0.2, 0.3, -0.1, -0.2, -0.3, 0.4])
k.setScale([1.5])
k.setAmplitude([2.0])

# We define a squared exponential kernel in the latent space as a reference
kRef = ot.SquaredExponential(3)
kRef.setScale(ot.Point(3, 1.5))
kRef.setAmplitude(ot.Point(1, 2.0))

ott.assert_almost_equal(
    k(1, 1)[0, 0], kRef(ot.Point([0.1, 0.0, 0.0]), ot.Point([0.1, 0.0, 0.0]))[0, 0]
)
ott.assert_almost_equal(
    k(1, 2)[0, 0], kRef(ot.Point([0.1, 0.0, 0.0]), ot.Point([0.2, 0.3, -0.1]))[0, 0]
)
ott.assert_almost_equal(
    k(0, 3)[0, 0], kRef(ot.Point([0.0, 0.0, 0.0]), ot.Point([-0.2, -0.3, 0.4]))[0, 0]
)



#! /usr/bin/env python

import openturns as ot
import os

ot.TESTPREAMBLE()

d = ot.Distribution()

# load
study = ot.Study()
study.setStorageManager(ot.XMLStorageManager("pyd.xml"))
study.load()
study.fillObject("d", d)

print(d.getKurtosis())
print(d.computePDF([0.5]))
print(d.computeCDF([0.5]))

os.remove("pyd.xml")



#! /usr/bin/env python

import openturns as ot
import os

ot.TESTPREAMBLE()

grids = [
    ot.RegularGrid(1.0, 0.1, 20),
    ot.RegularGrid(-3.0, 0.1, 20),
    ot.RegularGrid(3.0, -0.1, 20),
    ot.RegularGrid(-1.0, -0.1, 20),
    ot.RegularGrid(-1.0, 0.13, 20),
    ot.RegularGrid(1.0, -0.13, 20),
]

for regularGrid in grids:
    lowerBound = regularGrid.getLowerBound()[0]
    upperBound = regularGrid.getUpperBound()[0]
    n = regularGrid.getSimplicesNumber()
    print(
        "regularGrid=",
        regularGrid,
        "lowerBound=",
        lowerBound,
        "upperBound=",
        upperBound,
        n,
        "simplices",
    )
    algo = ot.RegularGridEnclosingSimplex(regularGrid)

    ot.RandomGenerator.SetSeed(0)
    test = ot.Sample(
        ot.Uniform(
            lowerBound - 0.2 * (upperBound - lowerBound),
            upperBound + 0.2 * (upperBound - lowerBound),
        ).getSample(1000)
    )

    vertices = regularGrid.getVertices()
    for vertex in test:
        index = algo.query(vertex)
        x = vertex[0]
        if x < lowerBound or x > upperBound:
            if index < n:
                print("Point", x, "should be outside but query returned index", index)
                os.exit(1)
        else:
            if index >= n:
                print("Point", x, "should be inside but query returned index", index)
                os.exit(1)

            found, coordinates = regularGrid.checkPointInSimplexWithCoordinates(
                vertex, index
            )
            if not found:
                print(
                    "Wrong simplex found for",
                    x,
                    "(index=",
                    index,
                    ") barycentric coordinates=",
                    coordinates,
                )
                os.exit(1)
            if (
                coordinates[0] < 0.0
                or coordinates[0] > 1.0
                or coordinates[1] < 0.0
                or coordinates[1] > 1.0
            ):
                print(
                    "Wrong barycentric coordinates found found for",
                    x,
                    "(index=",
                    index,
                    ") barycentric coordinates=",
                    coordinates,
                )
                os.exit(1)
            if (
                abs(
                    x
                    - coordinates[0] * vertices[index, 0]
                    - coordinates[1] * vertices[index + 1, 0]
                )
                > 1.0e-10
            ):
                print(
                    "Wrong barycentric coordinates found found for",
                    x,
                    "(index=",
                    index,
                    ") barycentric coordinates=",
                    coordinates,
                )
                os.exit(1)



"""
Estimate a multivariate distribution
====================================
"""

# %%
# In this example we are going to estimate a joint distribution from a multivariate sample by fitting marginals and finding a set of copulas.
#
# While the estimation of marginals is quite straightforward, the estimation of the dependency structure takes several steps:
#
# - find the dependent components
# - estimate a copula on each dependent bloc
# - assemble the estimated copulas
#

# %%
import openturns as ot
import math as m

# %%
# Generate some multivariate data to estimate, with correlation
cop1 = ot.AliMikhailHaqCopula(0.6)
cop2 = ot.ClaytonCopula(2.5)
copula = ot.BlockIndependentCopula([cop1, cop2])
marginals = [
    ot.Uniform(5.0, 6.0),
    ot.Arcsine(),
    ot.Normal(-40.0, 3.0),
    ot.Triangular(100.0, 150.0, 300.0),
]
distribution = ot.JointDistribution(marginals, copula)
sample = distribution.getSample(10000).getMarginal([0, 2, 3, 1])

# %%
# Estimate marginals
dimension = sample.getDimension()
marginalFactories = []
for factory in ot.DistributionFactory.GetContinuousUniVariateFactories():
    if str(factory).startswith("Histogram"):
        # ~ non-parametric
        continue
    marginalFactories.append(factory)
estimated_marginals = [
    ot.FittingTest.BestModelBIC(sample.getMarginal(i), marginalFactories)[0]
    for i in range(dimension)
]
estimated_marginals


# %%
# Find connected components of a graph defined from its adjacency matrix


# %%
def find_neighbours(head, covariance, to_visit, visited):
    visited[head] = 1
    to_visit.remove(head)
    current_component = [head]
    for i in to_visit:
        # If i is connected to head and has not yet been visited
        if covariance[head, i] > 0:
            # Add i to the current component
            component = find_neighbours(i, covariance, to_visit, visited)
            current_component += component
    return current_component


def connected_components(covariance):
    N = covariance.getDimension()
    to_visit = list(range(N))
    visited = [0] * N
    all_components = []
    for head in range(N):
        if visited[head] == 0:
            component = find_neighbours(head, covariance, to_visit, visited)
            all_components.append(sorted(component))
    return all_components


# %%
# Estimate the copula
# -------------------
#
# First find the dependent components : we compute the `Spearman` correlation

# %%
C = sample.computeSpearmanCorrelation()
print(C)

# %%
# We filter and consider only significantly non-zero correlations.

# %%
epsilon = 2.0 / m.sqrt(sample.getSize())
for j in range(dimension):
    for i in range(j):
        C[i, j] = 1.0 if abs(C[i, j]) > epsilon else 0.0
print(C)

# %%
# Note that we can apply the `HypothesisTest.Spearman` test. As the null hypothesis of the test is the `independence`, we must take the complementary of the binary measure as follow:
#
# >>>   M = ot.SymmetricMatrix(dimension)
# >>>   for i in range(dimension):
# >>>       M[i,i] = 1
# >>>       for j in range(i):
# >>>           M[i, j] = 1 - ot.HypothesisTest.Spearman(sample[:,i], sample[:,j]).getBinaryQualityMeasure()
#

# %%
# Now we find the independent blocs:

# %%
blocs = connected_components(C)
blocs

# %%
# For each dependent block, we estimate the most accurate parameteric copula.
#
# To do this, we first need to transform the sample in such a way as to keep the copula intact but make all marginal samples follow the uniform distribution on [0,1].

# %%
copula_sample = ot.Sample(sample.getSize(), sample.getDimension())
copula_sample.setDescription(sample.getDescription())
for index in range(sample.getDimension()):
    copula_sample[:, index] = estimated_marginals[index].computeCDF(sample[:, index])

# %%
copulaFactories = []
for factory in ot.DistributionFactory.GetContinuousMultiVariateFactories():
    if not factory.build().isCopula():
        continue
    if factory.getImplementation().getClassName() == "BernsteinCopulaFactory":
        continue
    copulaFactories.append(factory)
estimated_copulas = [
    ot.FittingTest.BestModelBIC(copula_sample.getMarginal(bloc), copulaFactories)[0]
    for bloc in blocs
]
estimated_copulas

# %%
# Finally we assemble the copula

# %%
estimated_copula_perm = ot.BlockIndependentCopula(estimated_copulas)

# %%
# Take care of the order of each bloc vs the order of original components !

# %%
permutation = []
for bloc in blocs:
    permutation.extend(bloc)
inverse_permutation = [-1] * dimension
for i in range(dimension):
    inverse_permutation[permutation[i]] = i
estimated_copula = estimated_copula_perm.getMarginal(inverse_permutation)
estimated_copula

# %%
# We build joint distribution from marginal distributions and dependency structure:

# %%
estimated_distribution = ot.JointDistribution(estimated_marginals, estimated_copula)
estimated_distribution



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


distribution = ot.AliMikhailHaqCopula(0.5)
size = 1000
sample = distribution.getSample(size)
factory = ot.AliMikhailHaqCopulaFactory()
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedDistribution = factory.build(distribution.getParameter())
print("Distribution from parameters=", estimatedDistribution)
estimatedAliMikhailHaqCopula = factory.buildAsAliMikhailHaqCopula(sample)
print("Distribution          =", distribution)
print("Estimated AliMikhailHaqCopula=", estimatedAliMikhailHaqCopula)
estimatedAliMikhailHaqCopula = factory.buildAsAliMikhailHaqCopula()
print("Default AliMikhailHaqCopula=", estimatedAliMikhailHaqCopula)
estimatedAliMikhailHaqCopula = factory.buildAsAliMikhailHaqCopula(
    distribution.getParameter()
)
print("AliMikhailHaqCopula from parameters=", estimatedAliMikhailHaqCopula)



import openturns as ot
import openturns.viewer as otv


class NormalTruncatedToBall(ot.PythonRandomVector):
    def __init__(self, dim, max_dist):
        super().__init__(dim)
        self._center = ot.Point(dim)
        self._normal = ot.Normal(dim)
        self._max_dist = max_dist
        self.setParameter(ot.Point(dim))

    def getRealization(self):
        dist = ot.SpecFunc.Infinity
        while dist > self._max_dist:
            candidate = self._normal.getRealization()
            dist = (candidate - self._center).norm()
        return candidate

    def setParameter(self, center):
        self._center = center

    def getParameter(self):
        return self._center

    def getParameterDescription(self):
        return [f"center_{i}" for i in range(self.getDimension())]


myRV = ot.RandomVector(NormalTruncatedToBall(2, 1.5))
myRV.setParameter([1.3, 0.6])
sample = myRV.getSample(1000)

graph = ot.Graph("Sample from a PythonRandomVector", "", "")
cloud = ot.Cloud(sample)
graph.add(cloud)
otv.View(graph)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# Generate the data for the polygons to be drawn
size = 50
cursor = ot.Point(2)

data1 = ot.Sample(size, 2)  # polygon y = 2x for x in [-25]
data2 = ot.Sample(size, 2)  # polygon y = x*x for x in [-11]

for i in range(size):
    tmp = 7.0 * i / size + 2
    cursor[0] = tmp
    cursor[1] = 2 * tmp
    data1[i] = cursor

    tmp = 9.0 * i / size + 1
    cursor[0] = tmp
    cursor[1] = tmp * tmp
    data2[i] = cursor

# Create an empty graph
graph = ot.Graph("Some polygons", "x1", "x2")
graph.setLegendPosition("upper right")

# Create the first polygon
polygon1 = ot.Polygon(data1)
polygon1.setColor("blue")

# Then, draw it
graph.add(polygon1)

# Create the second polygon
polygon2 = ot.Polygon(data2)
polygon2.setColor("red")

# Add it to the graph
graph.add(polygon2)

# Fill below the start of the second polygon
polygon = ot.Polygon.FillBetween(data2[:10, 0], data2[:10, 1], [[data2[0, 1]]] * 10)

polygon.setColor("green")
graph.add(polygon)

# Try all scale types
for i in range(4):
    graph.setLogScale(i)

print("graph = ", graph)



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

ot.Log.Show(ot.Log.WARN)

ll = [1.0, 0.7, 1.2, 0.9]
h = [0.5, 1.5, 3.5, 2.5]
distribution = ot.Histogram(-1.5, ll, h)
size = 10000
sample = distribution.getSample(size)
factory = ot.HistogramFactory()
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedHistogram = factory.buildAsHistogram(sample)
print("Histogram          =", distribution)
print("Estimated histogram=", estimatedHistogram)
estimatedHistogram = factory.buildAsHistogram(sample, 0.1)
print("Histogram          =", distribution)
print("Estimated histogram=", estimatedHistogram)
estimatedHistogram = factory.buildAsHistogram(sample, 15)
print("Histogram          =", distribution)
print("Estimated histogram=", estimatedHistogram)
first = -2.0
width = ot.Point(5, 1.0)
estimatedHistogram = factory.buildAsHistogram(sample, first, width)
print("Estimated histogram=", estimatedHistogram)
estimatedHistogram = factory.buildAsHistogram()
print("Default histogram=", estimatedHistogram)

sample = ot.ChiSquare(0.0120637).getSample(32)
for useQuantile in [True, False]:
    print(
        "useQuantile=",
        useQuantile,
        ot.Point(1, ot.HistogramFactory().computeBandwidth(sample, useQuantile)),
    )

# from quantiles
ref_dist = ot.Normal()
lowerBound = -3.0
probabilities = [
    0.1,
    0.2,
    0.3,
    0.4,
    0.5,
    0.6,
    0.7,
    0.8,
    0.9,
    1.0 - ot.SpecFunc.ScalarEpsilon,
]
quantiles = [ref_dist.computeQuantile(pi)[0] for pi in probabilities]
inf_distribution = ot.HistogramFactory().buildFromQuantiles(
    lowerBound, probabilities, quantiles
)
inf_quantiles = [inf_distribution.computeQuantile(pi)[0] for pi in probabilities]
ott.assert_almost_equal(inf_quantiles, quantiles)

# samples with different scales, bandwidth should not be too small and blow up memory thanks to MaximumBinNumber
sample1 = ot.Uniform(-1e-10, 1e-10).getSample(100)
sample2 = ot.Uniform(-1, 1).getSample(5)
sample1.add(sample2)
dist = ot.HistogramFactory().build(sample1)



"""
Create a linear combination of distributions
============================================
"""

# %%
import openturns as ot
import openturns.viewer as otv


# %%
# We define an affine combination of input random variables.
#
# .. math::
#    Y = 2 + 5 X_1 + X_2
#
# where:
#
# - :math:`X_1 \sim \mathcal{E}(\lambda=1.5)`
# - :math:`X_2 \sim \mathcal{N}(\mu=4, \sigma=1)`
#
# This notion is different from the mixture where the combination is made on the probability density functions and not on the univariate random variable.


# %%
# We create the distributions associated to the input random variables
X1 = ot.Exponential(1.5)
X2 = ot.Normal(4.0, 1.0)

# %%
# We define an offset `a0`
a0 = 2.0

# %%
# We create the `weights`
weight = [5.0, 1.0]

# %%
# We create the affine combination :math:`Y` :
distribution = ot.LinearCombinationDistribution([X1, X2], weight, a0)
print(distribution)

# %%
# We get its mean :
mean = distribution.getMean()[0]
print("Mean : %.3f" % mean)

# %%
# Its variance:
variance = distribution.getCovariance()[0, 0]
print("Variance : %.3f" % variance)

# %%
# The 90% quantile:
quantile = distribution.computeQuantile(0.9)[0]
print("0.9-quantile : %.3f" % quantile)

# %%
# We can get the probability of the :math:`Y` random variable to exceed 10.0 :
prb = distribution.computeSurvivalFunction(10.0)
print("Probability : %.3f" % prb)

# %%
# We draw its PDF :
graph = distribution.drawPDF()
view = otv.View(graph)

# %%
# We draw its CDF :
graph = distribution.drawCDF()
view = otv.View(graph)


# %%
# Create a discrete mixture
# -------------------------
#
# In this paragraph we build the distribution of the value of the sum of 20 dice rolls.
#
# .. math::
#    Y = \sum_{i=1}^{20} X_i
#
# where :math:`X_i \sim U(1,2,3,4,5,6)`
#

# %%
# We create the distribution associated to the dice roll :
X = ot.FiniteDiscreteDistribution([[i] for i in range(1, 7)])

# %%
# Let's roll the dice a few times !
sample = X.getSample(10)
print(sample)

# %%
N = 20

# %%
# We create a collection of identically distributed :math:`X_i`
coll = [X] * N

# %%
# We create the weights and an affine combination
weight = [1.0] * N
distribution = ot.LinearCombinationDistribution(coll, weight)

# %%
# We compute the probability to exceed a sum of 100 after 20 dice rolls
print("Probability : %.3g" % distribution.computeComplementaryCDF(100))

# %%
# We draw its PDF
graph = distribution.drawPDF()
view = otv.View(graph)

# %%
# and its CDF
graph = distribution.drawCDF()
view = otv.View(graph)

# %%
# Display all figures
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


class UniformNdPy(ot.PythonDistribution):
    def __init__(self, a=[-1.0], b=[1.0]):
        # https://github.com/uqfoundation/dill/issues/300
        # super().__init__(len(a))
        ot.PythonDistribution.__init__(self, len(a))
        if len(a) != len(b):
            raise ValueError("Invalid bounds")
        for i in range(len(a)):
            if a[i] > b[i]:
                raise ValueError("Invalid bounds")
        self.a = a
        self.b = b
        self.factor = 1.0
        for i in range(len(a)):
            self.factor *= b[i] - a[i]

    def getRange(self):
        return ot.Interval(self.a, self.b, [True] * len(self.a), [True] * len(self.a))

    def getRealization(self):
        X = []
        for i in range(len(self.a)):
            X.append(
                self.a[i] + (self.b[i] - self.a[i]) * ot.RandomGenerator.Generate()
            )
        return X

    def getSample(self, size):
        X = []
        for i in range(size):
            X.append(self.getRealization())
        return X

    def computeCDF(self, X):
        prod = 1.0
        for i in range(len(self.a)):
            if X[i] < self.a[i]:
                return 0.0
            prod *= min(self.b[i], X[i]) - self.a[i]
        return prod / self.factor

    def computePDF(self, X):
        for i in range(len(self.a)):
            if X[i] < self.a[i]:
                return 0.0
            if X[i] > self.b[i]:
                return 0.0
        return 1.0 / self.factor

    def getMean(self):
        mu = []
        for i in range(len(self.a)):
            mu.append(0.5 * (self.a[i] + self.b[i]))
        return mu


d = ot.Distribution(UniformNdPy())
print(d.getKurtosis())
print(d.computePDF([0.5]))
print(d.computeCDF([0.5]))

# save
study = ot.Study()
study.setStorageManager(ot.XMLStorageManager("pyd.xml"))
study.add("d", d)
study.save()



#! /usr/bin/env python

import openturns as ot

factory = ot.FourierSeriesFactory()
print(factory)
x = 0.4
for i in range(10):
    function = factory.build(i)
    print(
        "order=",
        i,
        function,
        "X=",
        ot.Point([x]),
        "f(X)=",
        ot.Point([function(x)]),
        "df(X)=",
        ot.Point([function.gradient(x)]),
        "d2f(X)=",
        ot.Point([function.hessian(x)]),
    )



#!/usr/bin/env python

import openturns as ot
import openturns.testing as ott

#
# Physical model
#

limitState = ot.SymbolicFunction(["u1", "u2"], ["u1-u2"])
dim = limitState.getInputDimension()

#
# Probabilistic model
#

mean = ot.Point(dim, 0.0)
mean[0] = 7.0
mean[1] = 2.0
sigma = ot.Point(dim, 1.0)

R = ot.IdentityMatrix(dim)
myDistribution = ot.Normal(mean, sigma, R)

#
# Limit state
#

vect = ot.RandomVector(myDistribution)

output = ot.CompositeRandomVector(limitState, vect)

myEvent = ot.ThresholdEvent(output, ot.Less(), 0.0)

#
# Computation
#
bs = 1

# Monte Carlo
experiment = ot.MonteCarloExperiment()
myMC = ot.ProbabilitySimulationAlgorithm(myEvent, experiment)
myMC.setMaximumOuterSampling(int(1e6) // bs)
myMC.setBlockSize(bs)
myMC.setMaximumCoefficientOfVariation(-1.0)
myMC.run()

#
# SubsetSampling
mySS = ot.SubsetSampling(myEvent)
mySS.setMaximumOuterSampling(10000 // bs)
mySS.setBlockSize(bs)
mySS.setKeepSample(True)
mySS.run()

#
# Results
#

# Monte Carlo
resultMC = myMC.getResult()
PFMC = resultMC.getProbabilityEstimate()
CVMC = resultMC.getCoefficientOfVariation()
variance_PF_MC = resultMC.getVarianceEstimate()
length90MC = resultMC.getProbabilityDistribution().computeBilateralConfidenceInterval(
    0.9
)
N_MC = resultMC.getOuterSampling() * resultMC.getBlockSize()

#
# SubsetSampling
resultSS = mySS.getResult()
PFSS = resultSS.getProbabilityEstimate()
CVSS = resultSS.getCoefficientOfVariation()
variance_PF_SS = resultSS.getVarianceEstimate()
length90SS = resultSS.getProbabilityDistribution().computeBilateralConfidenceInterval(
    0.9
)
N_SS = resultSS.getOuterSampling() * resultSS.getBlockSize()

#

print("-" * 100)
print("MONTE CARLO")
print("Pf estimation = %.5e" % PFMC)
print("Pf Variance estimation = %.5e" % variance_PF_MC)
print("CoV = %.5f" % CVMC)
print("90% Confidence Interval =", length90MC)
print("Limit state calls =", N_MC)
print("-" * 100)
print("SUBSET SAMPLING")
print("Pf estimation = %.5e" % PFSS)
print("Pf Variance estimation = %.5e" % variance_PF_SS)
print("CoV = %.5f" % CVSS)
print("90% Confidence Interval =", length90SS)
print("Limit state calls =", N_SS)


# check that the event sample is right
stepsNumber = mySS.getStepsNumber()
inputEventSample = mySS.getInputSample(stepsNumber - 1, mySS.EVENT1)
outputEventSample = mySS.getOutputSample(stepsNumber - 1, mySS.EVENT1)
outputG = limitState(inputEventSample)
diffSample = outputG - outputEventSample
ott.assert_almost_equal(diffSample.computeMean(), [0.0])

# null variance case
print("-" * 100)
f = ot.SymbolicFunction(["x"], ["x"])
X = ot.Normal()
Y = ot.CompositeRandomVector(f, ot.RandomVector(X))
event = ot.ThresholdEvent(Y, ot.Less(), 5000)
mc = ot.ProbabilitySimulationAlgorithm(event)
mc.run()
result = mc.getResult()
print("MC:    ", result)
subset = ot.SubsetSampling(event)
subset.setBlockSize(18)
subset.run()
result = subset.getResult()
print("SUBSET:", result)
assert subset.getStepsNumber() == 1, "wrong steps"
assert result.getProbabilityEstimate() == 1.0, "wrong pf"
assert result.getVarianceEstimate() == 0.0, "wrong var"

# case with last step threshold close from global threshold (before)
print("-" * 100)
event = ot.ThresholdEvent(Y, ot.Less(), -3.33832)
mc = ot.ProbabilitySimulationAlgorithm(event)
mc.run()
print("MC:    ", mc.getResult())
ot.RandomGenerator.SetSeed(65132)
subset = ot.SubsetSampling(event)
subset.setMaximumOuterSampling(500)
subset.run()
print("SUBSET:", subset.getResult())
print("steps=", subset.getStepsNumber())
print("T=", subset.getThresholdPerStep())

# case with last step threshold close from global threshold (after)
print("-" * 100)
event = ot.ThresholdEvent(Y, ot.Less(), -3.338325)
mc = ot.ProbabilitySimulationAlgorithm(event)
mc.run()
print("MC:    ", mc.getResult())
ot.RandomGenerator.SetSeed(65132)
subset = ot.SubsetSampling(event)
subset.setConvergenceStrategy(ot.Full())
subset.setMaximumOuterSampling(500)
subset.run()
print("SUBSET:", subset.getResult())
print("steps=", subset.getStepsNumber())
print("T=", subset.getThresholdPerStep())



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()
ot.PlatformInfo.SetNumericalPrecision(5)

ot.RandomGenerator.SetSeed(0)

# We create a Sample
point1 = [10.0, 20.0]
point2 = [11.0, 21.0]
point3 = [12.0, 22.0]
sample1 = [point1, point2, point3]
referencethreshold = [0, 3]

# Iterative threshold, one point at a time
dimension = 2
algo = ot.IterativeThresholdExceedance(dimension, ot.Greater(), 15.0)
algo.increment(point1)
algo.increment(point2)
algo.increment(point3)
computedthreshold = algo.getThresholdExceedance()
ott.assert_almost_equal(referencethreshold, computedthreshold)
iteration = algo.getIterationNumber()
ott.assert_almost_equal(iteration, 3)
algo.increment([20.0, 10.0])
ratio = algo.getRatio()
ott.assert_almost_equal(ratio, [0.25, 0.75])

# Iterative threshold, one single sample
iterthresholdSample = ot.IterativeThresholdExceedance(dimension, ot.Greater(), 15.0)
iterthresholdSample.increment(sample1)
computedthreshold = iterthresholdSample.getThresholdExceedance()
ott.assert_almost_equal(referencethreshold, computedthreshold)
iteration = iterthresholdSample.getIterationNumber()
ott.assert_almost_equal(iteration, 3)

# Iterative threshold, one single sample, then one point at a time
iterthresholdMixed = ot.IterativeThresholdExceedance(dimension, ot.Greater(), 15.0)
iterthresholdMixed.increment(sample1)
iterthresholdMixed.increment(point1)
iterthresholdMixed.increment(point2)
iterthresholdMixed.increment(point3)
computedthreshold = iterthresholdMixed.getThresholdExceedance()
referencethreshold = [0, 6]
ott.assert_almost_equal(referencethreshold, computedthreshold)
iteration = iterthresholdMixed.getIterationNumber()
ott.assert_almost_equal(iteration, 6)



#! /usr/bin/env python

import openturns as ot

ot.PlatformInfo.SetNumericalPrecision(6)

dim = 2

# First domain: [0,2]x[0,2]
cube = ot.Interval([0.0] * dim, [2.0] * dim)

# Second domain: sphere center=(0,0) r=1
function = ot.SymbolicFunction(["x", "y"], ["x^2 + y^2"])
sphere = ot.LevelSet(function, ot.LessOrEqual(), 1.0)

# Inside sphere but not cube
p0 = [-0.25, 0.25]
# Inside cube and sphere
p1 = [0.25, 0.25]
# Inside cube but not sphere
p2 = [1.8, 1.8]
# Outside
p3 = [4.0, 4.0]

domain = ot.DomainIntersection([cube, sphere])
print("cube=", cube)
print("sphere=", sphere)
print("intersection=", domain)
# Accessors
print("Dimension=", domain.getDimension())
# Contains
print("is point ", p0, " inside ? ", domain.contains(p0))
print("is point ", p1, " inside ? ", domain.contains(p1))
print("is point ", p2, " inside ? ", domain.contains(p2))
print("is point ", p3, " inside ? ", domain.contains(p3))
sample = [p0, p1, p2, p3]
print("is sample ", sample, " inside ? ", domain.contains(sample))

# p0 is outside cube, no need to check sphere2
sphere2 = ot.LevelSet(ot.SymbolicFunction(["x", "y"], ["x/0"]), ot.LessOrEqual(), 1.0)
domain2 = ot.DomainIntersection([cube, sphere2])
assert not domain2.contains(p0), "prune sphere"

# check thread-safety
f1 = ot.PythonFunction(2, 1, lambda X: [X[0]])
f2 = ot.PythonFunction(2, 1, lambda X: [X[1]])
l2 = ot.LevelSet(f2, ot.Greater(), 0.0)
l1 = ot.LevelSet(f1, ot.Greater(), 0.0)
u = ot.DomainIntersection([l1, l2])
x = ot.Normal(2).getSample(1000)
y = u.contains(x)
assert len(y) == 1000



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott
import math as m

v1 = [0.0, 0.0, 0.0]
v2 = [1.0, 1.0, 1.0]
v3 = [0.0, 1.0, 1.0]
v4 = [0.0, 0.0, 1.0]
# simplex
S = [v1, v2, v3, v4]
simplicies = [[0, 1, 2, 3]]
mesh = ot.Mesh(S, simplicies)
# function to integrate
f = ot.SymbolicFunction(["x1", "x2", "x3"], ["exp(x1 + x2 + x3)"])
# integral of f
algo = ot.SimplicialCubature()
value = algo.integrate(f, mesh)[0]
print(value)
ott.assert_almost_equal(value, (m.exp(1.0) - 1.0) ** 3 / 6)

# Test with interval inferface
f = ot.SymbolicFunction(["x", "y", "z"], ["sin(x) * cos(y) * exp(z)"])
valueRef = -m.sin(1.0) * (m.cos(1.0) - 1.0) * (m.e - 1.0)
value = algo.integrate(f, ot.Interval([0.0] * 3, [1.0] * 3))
ott.assert_almost_equal(value[0], valueRef)



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

dimension = 5
indices = [2, 0, 1]
coll = [ot.Normal(dimension), ot.Multinomial(10, [1.0 / (dimension + 2)] * dimension)]

for nDistribution in range(len(coll)):
    fullDistribution = coll[nDistribution]
    distribution = ot.MarginalDistribution(fullDistribution, indices)
    # To check if indices accessors are correct
    otherDistribution = distribution
    otherDistribution.setIndices(distribution.getIndices())
    print("Test indices accessors", otherDistribution == distribution)
    # To check if distribution accessors are correct
    otherDistribution.setDistribution(distribution.getDistribution())
    print("Test distribution accessors", otherDistribution == distribution)

    print("Distribution ", repr(distribution))
    print("Distribution ", distribution)

    # Is this distribution elliptical ?
    print("Elliptical = ", distribution.isElliptical())

    # Is this distribution continuous ?
    print("Continuous = ", distribution.isContinuous())

    # Is this distribution discrete ?
    print("Discrete = ", distribution.isDiscrete())

    # Is this distribution integral ?
    print("Integral = ", distribution.isIntegral())

    # Test for realization of distribution
    oneRealization = distribution.getRealization()
    print("oneRealization=", oneRealization)

    # Test for sampling
    size = 10000
    oneSample = distribution.getSample(size)
    print("oneSample first=", oneSample[0], " last=", oneSample[size - 1])
    print("mean=", oneSample.computeMean())
    print("covariance=", oneSample.computeCovariance())

    # Define a point
    point = [1.0] * distribution.getDimension()
    print("Point= ", point)

    # Test the DDF/logDDF only in the continuous case
    if distribution.isContinuous():
        DDF = distribution.computeDDF(point)
        print("ddf     =", DDF)
        LPDF = distribution.computeLogPDF(point)
        print("log pdf=%.5e" % LPDF)

    # Show PDF and CDF of point
    PDF = distribution.computePDF(point)
    print("pdf     =%.5e" % PDF)
    CDF = distribution.computeCDF(point)
    print("cdf     =%.5e" % CDF)
    CCDF = distribution.computeComplementaryCDF(point)
    print("ccdf    =%.5e" % CCDF)
    Survival = distribution.computeSurvivalFunction(point)
    print("survival=%.5e" % Survival)
    if distribution.isContinuous():
        InverseSurvival = distribution.computeInverseSurvivalFunction(0.95)
        print("Inverse survival=", InverseSurvival)
        print(
            "Survival(inverse survival)=%.5e"
            % distribution.computeSurvivalFunction(InverseSurvival)
        )
    quantile = distribution.computeQuantile(0.95)
    print("quantile=", quantile)
    print("cdf(quantile)=%.5e" % distribution.computeCDF(quantile))
    quantileTail = distribution.computeQuantile(0.95, True)
    print("quantile (tail)=", quantileTail)
    CDFTail = distribution.computeComplementaryCDF(quantileTail)
    print("cdf (tail)=%.5e" % CDFTail)
    mean = distribution.getMean()
    print("mean=", mean)
    standardDeviation = distribution.getStandardDeviation()
    print("standard deviation=", standardDeviation)
    skewness = distribution.getSkewness()
    print("skewness=", skewness)
    kurtosis = distribution.getKurtosis()
    print("kurtosis=", kurtosis)
    covariance = distribution.getCovariance()
    print("covariance=", covariance)
    correlation = distribution.getCorrelation()
    print("correlation=", correlation)
    if distribution.isContinuous():
        spearman = distribution.getSpearmanCorrelation()
        print("spearman=", spearman)
        kendall = distribution.getKendallTau()
        print("kendall=", kendall)
    print(
        "Standard representative=", distribution.getStandardRepresentative().__str__()
    )
    ot.Log.Show(ot.Log.TRACE)
    validation = ott.DistributionValidation(distribution)
    validation.skipDependenceMeasures()  # slow
    validation.skipConditional()  # FIXME
    validation.skipTransformation()  # FIXME
    validation.run()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

mesh = ot.RegularGrid(0.0, 0.1, 11)


def f(X):
    size = 11
    Y = [ot.Point(X) * i for i in range(size)]
    return Y


inputDim = 2
outputDim = 2
function = ot.PythonPointToFieldFunction(inputDim, mesh, outputDim, f)
function.setInputDescription(["u1", "u2"])
function.setOutputDescription(["v1", "v2"])

# freeze u1=5.0
parametric = ot.ParametricPointToFieldFunction(function, [0], [5.0])

# properties
print("dim=", parametric.getInputDimension(), parametric.getOutputDimension())
print(
    "description=", parametric.getInputDescription(), parametric.getOutputDescription()
)
print("mesh=", parametric.getOutputMesh())

# op(Point)
x = [1.0]
y = parametric(x)
print("y=", y)
print("calls=", parametric.getCallsNumber())


# op(Sample)
X = [[1.0], [7.0]]
Y = parametric(X)
print("Y=", Y)
print("calls=", parametric.getCallsNumber())



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott
from math import log10

ot.TESTPREAMBLE()

# Define the model
dimension = 3
input_names = ["x1", "x2", "x3"]
formulas = ["cos(x1 + x2 + x3)"]
model = ot.SymbolicFunction(input_names, formulas)

# Sample noisy data
distribution = ot.Normal(dimension)
samplesize = 100
x = distribution.getSample(samplesize)
y = model(x) + ot.Normal(0.0, 0.1).getSample(samplesize)

# Build the surrogate model
basis = ot.LinearBasisFactory(dimension).build()
covarianceModel = ot.SquaredExponential(dimension)
covarianceModel.activateNuggetFactor(True)

# Gaussian process fitter
fitter_algo = ot.GaussianProcessFitter(x, y, covarianceModel, basis)
fitter_algo.run()
fitter_result = fitter_algo.getResult()

algo = ot.GaussianProcessRegression(fitter_result)
algo.run()
result = algo.getResult()
optimized_covariance_model = result.getCovarianceModel()
ott.assert_almost_equal(
    optimized_covariance_model.getScale(), [0.874623, 0.491866, 0.84785], 0.2, 0.0
)
ott.assert_almost_equal(
    log10(optimized_covariance_model.getNuggetFactor()), -0.782689, 0.25, 0.0
)

# Product covariance kernel
# (with SquaredExponential => identical to the first)
product = ot.ProductCovarianceModel([ot.SquaredExponential()] * dimension)
product.activateNuggetFactor(True)

fitter_algo_product = ot.GaussianProcessFitter(x, y, product, basis)
fitter_algo_product.run()
fitter_result_product = fitter_algo_product.getResult()

algo_product = ot.GaussianProcessRegression(fitter_result_product)
algo_product.run()
result_product = algo_product.getResult()
optimized_product_model = result_product.getCovarianceModel()
ott.assert_almost_equal(
    log10(optimized_product_model.getNuggetFactor()), -0.782689, 0.25, 0.0
)

# Isotropic covariance kernel
isotropic = ot.IsotropicCovarianceModel(ot.SquaredExponential(), dimension)
isotropic.activateNuggetFactor(True)

fitter_algo_iso = ot.GaussianProcessFitter(x, y, isotropic, basis)
fitter_algo_iso.run()
fitter_result_iso = fitter_algo_iso.getResult()

algo_iso = ot.GaussianProcessRegression(fitter_result_iso)
algo_iso.run()
result_iso = algo_iso.getResult()
optimized_isotropic_model = result_iso.getCovarianceModel()
ott.assert_almost_equal(
    log10(optimized_isotropic_model.getNuggetFactor()), -1.9, 0.3, 0.0
)



"""
Create a white noise process
============================
"""

# %%
# This example details how to create and manipulate a white noise.
# A second order white noise
# :math:`\varepsilon: \Omega \times \mathbb{D} \rightarrow \mathbb{R}^d` is a stochastic
# process of dimension :math:`d` such that the covariance function
# :math:`C(\vect{s},\vect{t})=\delta(\vect{t}-\vect{s})C(\vect{s},\vect{s})`
# where :math:`C(\vect{s},\vect{s})` is the covariance matrix of the process at
# vertex :math:`\vect{s}` and :math:`\delta` the Kroenecker function.
#
# A process :math:`\varepsilon` is a white noise if all finite family of
# locations :math:`(\vect{t}_i)_{i=1, \dots, n} \in \mathbb{D}`,
# :math:`(\varepsilon_{\vect{t}_i})_{i=1, \dots, n}` is independent and
# identically distributed.
#
# The library proposes to model it through the object :class:`~openturns.WhiteNoise` defined
# on a mesh and a distribution with zero mean and finite standard
# deviation.
#
# If the distribution has a mean different from zero, The library writes
# message to prevent the User and does not allow the creation of such a
# white noise.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# Define the distribution
sigma = 1.0
dist = ot.Normal(0.0, sigma)

# %%
# Define the mesh
tgrid = ot.RegularGrid(0.0, 1.0, 100)

# %%
# Create the process
process = ot.WhiteNoise(dist, tgrid)
process

# %%
# Draw a realization
realization = process.getRealization()
graph = realization.drawMarginal(0)
graph.setTitle("Realization of a white noise with distribution N(0,1)")
view = otv.View(graph)

# %%
# Draw a sample
sample = process.getSample(5)
graph = sample.drawMarginal(0)
graph.setTitle(
    str(sample.getSize()) + " realizations of a white noise with distribution N(0,1)"
)
for k in range(sample.getSize()):
    drawable = graph.getDrawable(k)
    drawable.setLegend("realization " + str(k + 1))
    graph.setDrawable(k, drawable)
view = otv.View(graph)

# %%
# Display all figures
otv.View.ShowAll()



"""
=============================
Create multivariate functions
=============================
We can create multivariate functions by different methods. One of the methods is to gather
multivariate functions. The other one is to create a function having multivariate input using
the tensor product of univariate functions. In this example, we present both methods.
"""

# %%
# Description
# ===========
#
# We want to build some multivariate functions :math:`f : \mathbb{R}^d \mapsto \mathbb{R}^q`.
#
# We can do that:
#
# - Case 1: using a finite collection of multivariate functions :math:`f_i:  \mathbb{R}^d \mapsto \mathbb{R}^q`,
# - Case 2: using the tensor product of univariate functions,
# - Case 3: using the tensor product of orthogonal univariate polynomials,
# - Case 4: using the tensor product of orthogonal univariate functions.
#
#
# Case 1: Finite collection of multivariate functions
# ===================================================
#
# In that case, we have a  finite collection of multivariate functions
# :math:`f_i : \mathbb{R}^d \mapsto \mathbb{R}^q`.
# We use the :class:`~openturns.Basis` class.
#
# For example, we consider:
#
# .. math::
#     \begin{array}{lcl}
#       f_1(x_1, x_2) & = & (x_1^2, x_2^2)\\
#       f_2(x_1, x_2) & = & (x_1+1, 2x_2)\\
#       f_3(x_1, x_2) & = & (\cos(x_1x_2), x_2^3)
#     \end{array}

import openturns as ot

f1 = ot.SymbolicFunction(["x1", "x2"], ["x1^2", "x2^2"])
f2 = ot.SymbolicFunction(["x1", "x2"], ["x1+1.0", "2.0*x2"])
f3 = ot.SymbolicFunction(["x1", "x2"], ["cos(x1*x2)", "x2^3"])

myBasis = ot.Basis([f1, f2, f3])
f = myBasis.build(1)


# %%
# Case 2: Tensor product of univariate functions
# ==============================================
#
# In that case, the univariate functions are not necessarily orthogonal with respect to a measure.
#
# We have the family of functions: :math:`(x_i \mapsto \phi^i_k(x_i))_{k \geq 0}`
# for :math:`1 \leq i \leq d`.. The :class:`~openturns.EnumerateFunction` class
# stores the way the multivariate basis is enumerated: *enumerate(n)* is a multi-index
# :math:`\boldsymbol{\alpha} = (\alpha_1, \dots, \alpha_d)` such that the function number :math:`n` is:
#
# .. math::
#     f_n(\vect{x}) = \prod_{i=1}^d \phi^i_{\alpha_i}(x_i).
#
# We use the :class:`~openturns.TensorizedUniVariateFunctionFactory` class.
#
# For example, we consider: :math:`f: \mathbb{R}^2 \mapsto \mathbb{R}`, and:
#
# - dimension :math:`x_1`: the family of monomials: :math:`(x_1 \mapsto x_1^k)_{k \geq 0}`,
# - dimension :math:`x_2`: the family of Haar wavelets:  :math:`(x_2 \mapsto \phi^2_k(x_2))_{k \geq 0}`.

univFuncFamily_Mon = ot.MonomialFunctionFactory()
univFuncFamily_UnivPol = ot.OrthogonalUniVariatePolynomialFunctionFactory(
    ot.JacobiFactory()
)
univFuncFamily_Haar = ot.HaarWaveletFactory()
univFuncFamily_Fourier = ot.FourierSeriesFactory()

familyColl = [
    univFuncFamily_Mon,
    univFuncFamily_UnivPol,
    univFuncFamily_Haar,
    univFuncFamily_Fourier,
]
enumerateFunction = ot.LinearEnumerateFunction(len(familyColl))

familyFunction = ot.TensorizedUniVariateFunctionFactory(familyColl, enumerateFunction)
k = 3
f = familyFunction.build(k)
print("input dimension = ", f.getInputDimension())
print("output dimension = ", f.getOutputDimension())

# %%
# If we want to use an orthogonal univariate polynomials family, then we have to cast the family
# in the  :class:`~openturns.OrthogonalUniVariatePolynomialFunctionFactory` class.
#
# For example, we use the  Jacobi orthogonal univariate polynomials family.
univFuncFamily_Jacobi = ot.OrthogonalUniVariatePolynomialFunctionFactory(
    ot.JacobiFactory()
)


# %%
# Case 3:  Tensor product of orthogonal univariate polynomials
# ============================================================
#
# In that case, the univariate polynomials are orthogonal with respect to a measure :math:`\mu`.
#
# We have the family of polynomials: :math:`(x_i \mapsto \phi^i_k(x_i))_{k \geq 0}`
# for :math:`1 \leq i \leq d`. The :class:`~openturns.EnumerateFunction` class decides the way
# the multivariate basis is enumerated: *enumerate(n)* is a multi-index
# :math:`(\ell_1, \dots, \ell_d)` such that the function number :math:`n`
# is:
#
# .. math::
#     f_n(\vect{x}) = \prod_{i=1}^d \phi^i_{\ell_i}(x_i).
#
# We use the  :class:`~openturns.OrthogonalProductPolynomialFactory` class.
#
# For example, we consider : :math:`f: \mathbb{R}^2 \mapsto \mathbb{R}`, and:
#
# - dimension of :math:`x_1`: the family of Jacobi polynomials: :math:`(x_1 \mapsto x_1^k){k \geq 0}`,
# - dimension of :math:`x_2`: the family of Hermite polynomials: :math:`(x_2 \mapsto \phi^2_k(x_2))_{k \geq 0}`.

univFuncFamily_Jacobi = ot.JacobiFactory()
univFuncFamily_Hermite = ot.HermiteFactory()
familyColl = [univFuncFamily_Jacobi, univFuncFamily_Hermite]
enumerateFunction = ot.LinearEnumerateFunction(len(familyColl))
familyFunction = ot.OrthogonalProductPolynomialFactory(familyColl, enumerateFunction)
f = familyFunction.build(3)
print("input dimension = ", f.getInputDimension())
print("output dimension = ", f.getOutputDimension())

# %%
# We get the measure:
measure_Jacobi = ot.JacobiFactory().getMeasure()
measure_Hermite = ot.HermiteFactory().getMeasure()
print("Measure orthogonal to Jacobi polynomials = ", measure_Jacobi)
print("Measure orthogonal to Hermite polynomials = ", measure_Hermite)


# %%
# Case 4: Tensor product of orthogonal univariate functions
# =========================================================
#
# In that case, the univariate functions are orthogonal with respect to a measure :math:`\mu`.
#
# We have the family of functions: :math:`(x_i \mapsto \phi^i_k(x_i))_{k \geq 0}`
# for :math:`1 \leq i \leq d`. The :class:`~openturns.EnumerateFunction` class
# stores the way the multivariate basis is enumerated: *enumerate(n)* is a multi-index
# :math:`\vect{\alpha} = (\alpha_1, \dots, \alpha_d)` such that the function number *n* is:
#
# .. math::
#
#      f_n(\vect{x}) = \prod_{i=1}^d \phi^i_{\alpha_i}(x_i)
#
# We use the :class:`~openturns.OrthogonalProductFunctionFactory` class.


# %%
# For example, we consider : :math:`f: \mathbb{R}^2 \mapsto \mathbb{R}`, and:
#
# - dimension :math:`x_1`: the family of Haar wavelets: :math:`(x_1 \mapsto \phi^1_k(x_1))_{k \geq 0}`,
# - dimension :math:`x_2`: the family of Fourier series functions: :math:`(x_2 \mapsto \phi^2_k(x_2)){k \geq 0}`.

univFuncFamily_Haar = ot.HaarWaveletFactory()
univFuncFamily_Fourier = ot.FourierSeriesFactory()
familyColl = [univFuncFamily_Haar, univFuncFamily_Fourier]
enumerateFunction = ot.LinearEnumerateFunction(len(familyColl))

familyFunction = ot.OrthogonalProductFunctionFactory(familyColl, enumerateFunction)
k = 3
f = familyFunction.build(k)

# %%
# We get the measure:
measure_Haar = ot.HaarWaveletFactory().getMeasure()
measure_Fourier = ot.FourierSeriesFactory().getMeasure()
print("Measure orthogonal to Haar wavelets = ", measure_Haar)
print("Measure orthogonal to Fourier series = ", measure_Fourier)
